from __future__ import annotations

from datetime import timedelta
from typing import Any

import pydantic
from prefect import flow, get_run_logger
from prefect.automations import AutomationCore
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.client.schemas.objects import State  # noqa: TC002
from prefect.events.actions import RunDeployment
from prefect.events.schemas.automations import EventTrigger, Posture
from prefect.states import Completed, Failed

from infrahub import lock
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.merge import BranchMerger
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.branch_action import BranchCreateEvent, BranchDeleteEvent, BranchRebaseEvent
from infrahub.exceptions import BranchNotFoundError, MergeFailedError, ValidationError
from infrahub.graphql.mutations.models import BranchCreateModel  # noqa: TC001
from infrahub.log import get_log_data
from infrahub.message_bus import Meta, messages
from infrahub.services import services
from infrahub.worker import WORKER_IDENTITY
from infrahub.workflows.catalogue import (
    BRANCH_CANCEL_PROPOSED_CHANGES,
    COMPUTED_ATTRIBUTE_REMOVE_PYTHON,
    COMPUTED_ATTRIBUTE_SETUP_PYTHON,
    DIFF_REFRESH_ALL,
    GIT_REPOSITORIES_CREATE_BRANCH,
    IPAM_RECONCILIATION,
)
from infrahub.workflows.utils import add_branch_tag

from .constants import AUTOMATION_NAME_CREATE, AUTOMATION_NAME_REMOVE


@flow(name="branch-rebase", flow_run_name="Rebase branch {branch}")
async def rebase_branch(branch: str) -> None:
    service = services.service

    async with service.database.start_session() as db:
        log = get_run_logger()
        await add_branch_tag(branch_name=branch)
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            service=service,
        )
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch that must be addressed."
                " Please review the diff for details and manually update the conflicts before rebasing."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)

        # If there are some changes related to the schema between this branch and main, we need to
        #  - Run all the validations to ensure everything is correct before rebasing the branch
        #  - Run all the migrations after the rebase
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)
        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(branch=obj, schema_branch=candidate_schema, constraints=constraints)
            )
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        schema_in_main_before = merger.destination_schema.duplicate()

        async with lock.registry.global_graph_lock():
            async with db.start_transaction() as dbt:
                await obj.rebase(db=dbt)
                log.info("Branch successfully rebased")

            if obj.has_schema_changes:
                # NOTE there is a bit additional work in order to calculate a proper diff that will
                # allow us to pull only the part of the schema that has changed, for now the safest option is to pull
                # Everything
                # schema_diff = await merger.has_schema_changes()
                # TODO Would be good to convert this part to a Prefect Task in order to track it properly
                updated_schema = await registry.schema.load_schema_from_db(
                    db=db,
                    branch=obj,
                    # schema=merger.source_schema.duplicate(),
                    # schema_diff=schema_diff,
                )
                registry.schema.set_schema_branch(name=obj.name, schema=updated_schema)
                obj.update_schema_hash()
                await obj.save(db=db)

                # Execute the migrations
                migrations = await merger.calculate_migrations(target_schema=updated_schema)

                errors = await schema_apply_migrations(
                    message=SchemaApplyMigrationData(
                        branch=merger.source_branch,
                        new_schema=candidate_schema,
                        previous_schema=schema_in_main_before,
                        migrations=migrations,
                    )
                )
                for error in errors:
                    log.error(error)

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the rebase
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION, parameters={"branch": obj.name, "ipam_node_details": ipam_node_details}
            )

    await service.workflow.submit_workflow(workflow=DIFF_REFRESH_ALL, parameters={"branch_name": obj.name})

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been rebased
    # -------------------------------------------------------------
    await service.event.send(event=BranchRebaseEvent(branch=obj.name, branch_id=obj.get_id()))


@flow(name="branch-merge", flow_run_name="Merge branch {branch} into main")
async def merge_branch(branch: str) -> None:
    service = services.service
    async with service.database.start_session() as db:
        log = get_run_logger()

        await add_branch_tag(branch_name=branch)
        await add_branch_tag(branch_name=registry.default_branch)

        obj = await Branch.get_by_name(db=db, name=branch)
        component_registry = get_component_registry()

        merger: BranchMerger | None = None
        async with lock.registry.global_graph_lock():
            # await update_diff(model=RequestDiffUpdate(branch_name=obj.name))

            diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
            diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
            diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
            merger = BranchMerger(
                db=db,
                diff_coordinator=diff_coordinator,
                diff_merger=diff_merger,
                diff_repository=diff_repository,
                source_branch=obj,
                service=service,
            )
            try:
                await merger.merge()
            except Exception as exc:
                log.exception("Merge failed, beginning rollback")
                await merger.rollback()
                raise MergeFailedError(branch_name=branch) from exc
            await merger.update_schema()

        if merger and merger.migrations:
            errors = await schema_apply_migrations(
                message=SchemaApplyMigrationData(
                    branch=merger.destination_branch,
                    new_schema=merger.destination_schema,
                    previous_schema=merger.initial_source_schema,
                    migrations=merger.migrations,
                )
            )
            for error in errors:
                log.error(error)

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the merge
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await service.workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                parameters={"branch": registry.default_branch, "ipam_node_details": ipam_node_details},
            )
        # -------------------------------------------------------------
        # remove tracking ID from the diff because there is no diff after the merge
        # -------------------------------------------------------------
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        await diff_repository.mark_tracking_ids_merged(tracking_ids=[BranchTrackingId(name=obj.name)])

        # -------------------------------------------------------------
        # Generate an event to indicate that a branch has been merged
        # NOTE: we still need to convert this event and potentially pull
        #   some tasks currently executed based on the event into this workflow
        # -------------------------------------------------------------
        log_data = get_log_data()
        request_id = log_data.get("request_id", "")
        message = messages.EventBranchMerge(
            source_branch=obj.name,
            target_branch=registry.default_branch,
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=request_id),
        )
        await service.send(message=message)


@flow(name="branch-delete", flow_run_name="Delete branch {branch}")
async def delete_branch(branch: str) -> None:
    service = services.service

    await add_branch_tag(branch_name=branch)

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=str(branch))
        event = BranchDeleteEvent(branch=branch, branch_id=obj.get_id(), sync_with_git=obj.sync_with_git)
        await obj.delete(db=db)

        await service.workflow.submit_workflow(
            workflow=BRANCH_CANCEL_PROPOSED_CHANGES, parameters={"branch_name": branch}
        )

        await service.event.send(event=event)


@flow(
    name="branch-validate",
    flow_run_name="Validate branch {branch} for conflicts",
    description="Validate if the branch has some conflicts",
    persist_result=True,
)
async def validate_branch(branch: str) -> State:
    service = services.service
    await add_branch_tag(branch_name=branch)

    async with service.database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        has_conflicts = await diff_repo.diff_has_conflicts(
            diff_branch_name=obj.name, tracking_id=BranchTrackingId(name=obj.name)
        )
        if has_conflicts:
            return Failed(message="branch has some conflicts")
        return Completed(message="branch is valid")


@flow(name="create-branch", flow_run_name="Create branch {model.name}")
async def create_branch(model: BranchCreateModel) -> None:
    service = services.service
    await add_branch_tag(model.name)

    async with service.database.start_session() as db:
        try:
            await Branch.get_by_name(db=db, name=model.name)
            raise ValueError(f"The branch {model.name}, already exist")
        except BranchNotFoundError:
            pass

        data_dict: dict[str, Any] = dict(model)
        if "is_isolated" in data_dict:
            del data_dict["is_isolated"]

        try:
            obj = Branch(**data_dict)
        except pydantic.ValidationError as exc:
            error_msgs = [f"invalid field {error['loc'][0]}: {error['msg']}" for error in exc.errors()]
            raise ValueError("\n".join(error_msgs)) from exc

        async with lock.registry.local_schema_lock():
            # Copy the schema from the origin branch and set the hash and the schema_changed_at value
            origin_schema = registry.schema.get_schema_branch(name=obj.origin_branch)
            new_schema = origin_schema.duplicate(name=obj.name)
            registry.schema.set_schema_branch(name=obj.name, schema=new_schema)
            obj.update_schema_hash()
            await obj.save(db=db)

            # Add Branch to registry
            registry.branch[obj.name] = obj
            await service.component.refresh_schema_hash(branches=[obj.name])

        event = BranchCreateEvent(branch=obj.name, branch_id=str(obj.uuid), sync_with_git=obj.sync_with_git)
        await service.event.send(event=event)

        if obj.sync_with_git:
            await service.workflow.submit_workflow(
                workflow=GIT_REPOSITORIES_CREATE_BRANCH,
                parameters={"branch": obj.name, "branch_id": str(obj.uuid)},
            )


@flow(name="branch-actions-setup", flow_run_name="Setup branch action events in task-manager")
async def branch_actions_setup() -> None:
    log = get_run_logger()

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(
                        any_=[COMPUTED_ATTRIBUTE_SETUP_PYTHON.name, COMPUTED_ATTRIBUTE_REMOVE_PYTHON.name]
                    )
                )
            )
        }
        deployment_id_computed_attribute_setup_python = deployments[COMPUTED_ATTRIBUTE_SETUP_PYTHON.name].id
        deployment_id_computed_attribute_remove_python = deployments[COMPUTED_ATTRIBUTE_REMOVE_PYTHON.name].id

        branch_create_automation = await client.find_automation(id_or_name=AUTOMATION_NAME_CREATE)

        automation = AutomationCore(
            name=AUTOMATION_NAME_CREATE,
            description="Trigger actions on branch create event",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.branch.created"},
                within=timedelta(0),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_computed_attribute_setup_python,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                        "trigger_updates": False,
                    },
                    job_variables={},
                ),
            ],
        )

        if branch_create_automation:
            await client.update_automation(automation_id=branch_create_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME_CREATE} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME_CREATE} Created")

        branch_remove_automation = await client.find_automation(id_or_name=AUTOMATION_NAME_REMOVE)

        automation = AutomationCore(
            name=AUTOMATION_NAME_REMOVE,
            description="Trigger actions on branch delete event",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.branch.deleted"},
                within=timedelta(0),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_computed_attribute_remove_python,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    },
                    job_variables={},
                ),
            ],
        )

        if branch_remove_automation:
            await client.update_automation(automation_id=branch_remove_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME_REMOVE} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME_REMOVE} Created")



from __future__ import annotations

from prometheus_client import Counter, Histogram

METRIC_PREFIX = "infrahub_db"

QUERY_EXECUTION_METRICS = Histogram(
    f"{METRIC_PREFIX}_query_execution_seconds",
    "Execution time to query the database",
    labelnames=["type", "query", "runtime", "context1", "context2"],
    buckets=[0.005, 0.01, 0.02, 0.03, 0.04, 0.05, 0.1, 0.5, 1],
)

TRANSACTION_RETRIES = Counter(
    f"{METRIC_PREFIX}_transaction_retries",
    "Number of transaction that have been retried due to transcient error",
    labelnames=["name"],
)



from pathlib import Path
from typing import Any, Optional, Union


class Error(Exception):
    HTTP_CODE: int = 500
    DESCRIPTION: str = "Unknown Error"
    message: str = ""
    errors: Optional[list] = None

    def api_response(self) -> dict[str, Any]:
        """Return error response."""
        if isinstance(self.errors, list):
            return {"data": None, "errors": self.errors}
        return {
            "data": None,
            "errors": [{"message": str(self.message) or self.DESCRIPTION, "extensions": {"code": self.HTTP_CODE}}],
        }


class PropagatedFromWorkerError(Error):
    """
    Used to re-raise server side an error that happened worker side.
    Note we might want to improve this so we raise the exact same error that happened worker side.
    """

    def __init__(self, http_code: int, message: str) -> None:
        self.HTTP_CODE = http_code
        self.message = message


class RPCError(Error):
    HTTP_CODE: int = 502

    def __init__(self, message: str) -> None:
        self.message = message


class InitializationError(Error):
    DESCRIPTION: str = "The application hasn't been initialized properly"


class DatabaseError(Error):
    HTTP_CODE: int = 503
    DESCRIPTION = "Database unavailable"

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class LockError(Error):
    pass


class GraphQLQueryError(Error):
    HTTP_CODE = 502

    def __init__(self, errors: list) -> None:
        self.errors = errors


class RepositoryError(Error):
    def __init__(self, identifier: str, message: str | None = None) -> None:
        self.identifier = identifier
        self.message = message or f"An error occurred with GitRepository '{identifier}'."
        super().__init__(self.message)


class RepositoryInvalidFileSystemError(RepositoryError):
    def __init__(
        self,
        identifier: str,
        directory: Path,
        message: str | None = None,
    ) -> None:
        super().__init__(identifier=identifier)
        self.directory = directory
        self.message = message or f"Invalid file system for {identifier}, Local directory {directory} missing."


class CommitNotFoundError(Error):
    HTTP_CODE: int = 400

    def __init__(self, identifier: str, commit: str, message: str | None = None) -> None:
        self.identifier = identifier
        self.commit = commit
        self.message = message or f"Commit {commit} not found with GitRepository '{identifier}'."
        super().__init__(self.message)


class DataTypeNotFoundError(Error):
    HTTP_CODE: int = 400

    def __init__(self, name: str, message: str | None = None) -> None:
        self.name = name
        self.message = message or f"Unable to find the DataType '{name}'."
        super().__init__(self.message)


class RepositoryFileNotFoundError(Error):
    HTTP_CODE: int = 404

    def __init__(self, repository_name: str, location: str, commit: str, message: str | None = None) -> None:
        self.repository_name = repository_name
        self.location = location
        self.commit = commit
        self.message = message or f"Unable to find the file at '{repository_name}::{commit}::{location}'."
        super().__init__(self.message)


class FileOutOfRepositoryError(Error):
    HTTP_CODE: int = 403

    def __init__(self, repository_name: str, location: str, commit: str, message: str | None = None) -> None:
        self.repository_name = repository_name
        self.location = location
        self.commit = commit
        self.message = message or f"File not in repository '{repository_name}::{commit}::{location}'."
        super().__init__(self.message)


class TransformError(Error):
    def __init__(self, repository_name: str, location: str, commit: str, message: str | None = None) -> None:
        self.repository_name = repository_name
        self.location = location
        self.commit = commit
        self.message = (
            message or f"An error occurred with the transform function at '{repository_name}::{commit}::{location}'."
        )
        super().__init__(self.message)


class CheckError(Error):
    def __init__(
        self, repository_name: str, location: str, class_name: str, commit: str, message: str | None = None
    ) -> None:
        self.repository_name = repository_name
        self.location = location
        self.commit = commit
        self.class_name = class_name
        self.message = (
            message
            or f"An error occurred with the check function at '{repository_name}::{commit}::{location}::{class_name}'."
        )
        super().__init__(self.message)


class TransformNotFoundError(TransformError):
    def __init__(self, repository_name: str, location: str, commit: str, message: str | None = None) -> None:
        self.message = (
            message or f"Unable to locate the transform function at '{repository_name}::{commit}::{location}'."
        )
        super().__init__(repository_name, location, commit, self.message)


class BranchNotFoundError(Error):
    HTTP_CODE: int = 400

    def __init__(self, identifier: str, message: str | None = None) -> None:
        self.identifier = identifier
        self.message = message or f"Branch: {identifier} not found."
        super().__init__(self.message)


class NodeNotFoundError(Error):
    HTTP_CODE: int = 404

    def __init__(
        self, node_type: str, identifier: str, branch_name: str | None = None, message: str | None = None
    ) -> None:
        self.node_type = node_type
        self.identifier = identifier
        self.branch_name = branch_name
        self.message = message or f"Unable to find the node {identifier} / {node_type} in the database."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"""
        {self.message}
        {self.branch_name} | {self.node_type} | {self.identifier}
        """


class ResourceNotFoundError(Error):
    HTTP_CODE: int = 404

    def __init__(self, message: str | None = None) -> None:
        self.message = message or "The requested resource was not found"
        super().__init__(self.message)


class AuthorizationError(Error):
    HTTP_CODE: int = 401
    message: str = "Access to the requested resource was denied"

    def __init__(self, message: str | None = None) -> None:
        self.message = message or self.message
        super().__init__(self.message)


class PermissionDeniedError(Error):
    HTTP_CODE: int = 403
    message: str = "The requested operation was not authorized"

    def __init__(self, message: str | None = None) -> None:
        self.message = message or self.message
        super().__init__(self.message)


class ProcessingError(Error):
    HTTP_CODE: int = 400
    message: str = "Unable to process the request"

    def __init__(self, message: str | None = None) -> None:
        self.message = message or self.message
        super().__init__(self.message)


class PoolExhaustedError(Error):
    HTTP_CODE: int = 409
    message: str = "No more resources available in the pool"

    def __init__(self, message: str | None = None) -> None:
        self.message = message or self.message
        super().__init__(self.message)


class SchemaNotFoundError(Error):
    HTTP_CODE: int = 422

    def __init__(self, branch_name: str, identifier: str, message: str | None = None) -> None:
        self.branch_name = branch_name
        self.identifier = identifier
        self.message = message or f"Unable to find the schema {identifier} in the database."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"""
        {self.message}
        {self.branch_name} | {self.identifier}
        """


class QueryError(Error):
    def __init__(self, query: str, params: dict, message: str = "Unable to execute the CYPHER query.") -> None:
        self.query = query
        self.params = params

        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"""
        {self.message}
        {self.query}
        {self.params}
        """


class QueryValidationError(Error):
    HTTP_CODE = 400

    def __init__(self, message: str) -> None:
        self.message = message


class GatewayError(Error):
    HTTP_CODE = 502

    def __init__(self, message: str) -> None:
        self.message = message


class MigrationError(Error):
    HTTP_CODE = 502

    def __init__(self, message: str) -> None:
        self.message = message


class ValidationError(Error):
    HTTP_CODE = 422

    def __init__(self, input_value: Union[str, dict, list]) -> None:
        self.message = ""
        self.location = None
        self.messages = {}

        if isinstance(input_value, str):
            self.message = input_value
        elif isinstance(input_value, dict) and len(input_value) == 1:
            self.message = list(input_value.values())[0]
            self.location = list(input_value.keys())[0]
        elif isinstance(input_value, dict) and len(input_value) > 1:
            for key, value in input_value.items():
                self.messages[key] = value

        elif isinstance(input_value, list):
            for item in input_value:
                if isinstance(item, self.__class__):
                    self.messages[item.location] = item.message
                elif isinstance(item, dict):
                    for key, value in item.items():
                        self.messages[key] = value

        super().__init__(self.message)

    def __str__(self) -> str:
        if self.messages:
            return ", ".join([f"{message} at {location}" for location, message in self.messages.items()])

        return f"{self.message} at {self.location or '<Undefined>'}"


class DiffError(Error):
    HTTP_CODE = 400

    def __init__(self, message: str) -> None:
        self.message = message


class DiffRangeValidationError(DiffError): ...


class DiffFromRequiredOnDefaultBranchError(DiffError): ...


class HTTPServerError(Error):
    """Errors raised when communicating with external HTTP servers"""

    HTTP_CODE = 502

    def __init__(self, message: str) -> None:
        self.message = message


class HTTPServerTimeoutError(HTTPServerError):
    HTTP_CODE = 504


class HTTPServerSSLError(HTTPServerError):
    HTTP_CODE = 503


class MergeFailedError(Error):
    HTTP_CODE: int = 500

    def __init__(self, branch_name: str) -> None:
        self.message = f"Failed to merge branch '{branch_name}'"
        super().__init__(self.message)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import Severity, TaskConclusion
from infrahub.log import get_logger

from .task import Task
from .task_log import TaskLog

if TYPE_CHECKING:
    from types import TracebackType

    from structlog.stdlib import BoundLogger

    from infrahub.core.protocols import CoreGenericAccount
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext
    from infrahub.services.protocols import InfrahubLogger


class UserTask:
    def __init__(
        self,
        title: str,
        db: InfrahubDatabase,
        account: Optional[CoreGenericAccount] = None,
        account_id: Optional[str] = None,
        logger: Optional[Union[BoundLogger, InfrahubLogger]] = None,
    ) -> None:
        if not account and not account_id:
            raise ValueError("Either account or account_id must be provided to initialize UserTask")

        self._account = account
        if account_id:
            self.account_id = account_id
        if not account_id and account:
            self.account_id = account.id

        self.title = title
        self._task: Optional[Task]
        self.log: Union[BoundLogger, InfrahubLogger] = logger or get_logger()
        self.db = db

    @property
    def task(self) -> Task:
        if self._task:
            return self._task
        raise ValueError("Task hasn't been initialized")

    @property
    def account(self) -> CoreGenericAccount:
        if self._account:
            return self._account
        raise ValueError("Account hasn't been initialized")

    @property
    def task_id(self) -> str:
        if self.task.uuid:
            return str(self.task.uuid)
        raise ValueError("Task hasn't been initialized")

    async def fetch_account(self) -> bool:
        if self._account:
            return False

        account: Optional[CoreGenericAccount] = await registry.manager.get_one(id=self.account_id, db=self.db)
        if not account:
            raise ValueError(f"Unable to find the account associated with {self.account_id}")
        self._account = account
        return True

    async def create_task(self) -> None:
        await self.fetch_account()
        self._task = Task(title=self.title, conclusion=TaskConclusion.UNKNOWN, related_node=self.account)
        await self._task.save(db=self.db)

    @classmethod
    def from_graphql_context(
        cls, title: str, context: GraphqlContext, logger: Optional[Union[BoundLogger, InfrahubLogger]] = None
    ) -> Self:
        if not context.db or not context.account_session:
            raise ValueError("db and account_session must be provided to initialize a GraphQLTaskReport")

        if not logger and context.service and context.service.log:
            logger = context.service.log
        return cls(title=title, account_id=context.account_session.account_id, db=context.db, logger=logger)

    async def __aenter__(self) -> Self:
        await self.create_task()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_type:
            await self.error(message=str(exc_value))
            self.task.conclusion = TaskConclusion.FAILURE
        else:
            self.task.conclusion = TaskConclusion.SUCCESS
        await self.task.save(db=self.db)

    async def add_log(
        self, message: str, severity: Severity = Severity.INFO, db: Optional[InfrahubDatabase] = None
    ) -> None:
        tlog = TaskLog(message=message, severity=severity, task_id=self.task_id)
        await tlog.save(db=db or self.db)

    async def info(self, message: str, db: Optional[InfrahubDatabase] = None, **kwargs: Any) -> None:
        if self.log:
            self.log.info(message, **kwargs)
        await self.add_log(message=message, severity=Severity.INFO, db=db)

    async def warning(self, message: str, db: Optional[InfrahubDatabase] = None, **kwargs: Any) -> None:
        if self.log:
            self.log.warning(message, **kwargs)
        await self.add_log(message=message, severity=Severity.WARNING, db=db)

    async def error(self, message: str, db: Optional[InfrahubDatabase] = None, **kwargs: Any) -> None:
        if self.log:
            self.log.error(message, **kwargs)
        await self.add_log(message=message, severity=Severity.ERROR, db=db)



from typing import Any

from infrahub.core.branch.models import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class ArtifactDiffQuery(Query):
    name = "get_artifact_diff"
    type = QueryType.READ

    def __init__(
        self,
        target_branch: Branch,
        target_rel_identifier: str,
        definition_rel_identifier: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.target_branch = target_branch
        self.target_rel_identifier = target_rel_identifier
        self.definition_rel_identifier = definition_rel_identifier

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        source_branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params.update(
            {
                "source_branch_name": self.branch.name,
                "target_branch_name": self.target_branch.name,
                "target_rel_identifier": self.target_rel_identifier,
                "definition_rel_identifier": self.definition_rel_identifier,
            }
        )
        query = """
// -----------------------
// get the active artifacts on the source branch
// -----------------------
MATCH (source_artifact:%(artifact_kind)s)-[r:IS_PART_OF]->(:Root)
WHERE r.branch IN [$source_branch_name, $target_branch_name]
CALL {
    WITH source_artifact
    MATCH (source_artifact)-[r:IS_PART_OF]->(:Root)
    WHERE %(source_branch_filter)s
    RETURN r AS root_rel
    ORDER BY r.branch_level DESC, r.from DESC
    LIMIT 1
}
WITH source_artifact, root_rel
WHERE root_rel.status = "active"
CALL {
    WITH source_artifact
    // -----------------------
    // get the artifact's target node
    // -----------------------
    CALL {
        WITH source_artifact
        OPTIONAL MATCH (source_artifact)-[rrel1:IS_RELATED]-(rel_node:Relationship)-[rrel2:IS_RELATED]-(target_node:Node)
        WHERE rel_node.name = $target_rel_identifier
        AND all(r IN [rrel1, rrel2] WHERE ( %(source_branch_filter)s ))
        RETURN
            target_node,
            (rrel1.status = "active" AND rrel2.status = "active") AS target_is_active,
            $source_branch_name IN [rrel1.branch, rrel2.branch] AS target_on_source_branch
        ORDER BY rrel1.branch_level DESC, rrel1.branch_level DESC, rrel1.from DESC, rrel2.from DESC
        LIMIT 1
    }
    // -----------------------
    // get the artifact's definition node
    // -----------------------
    CALL {
        WITH source_artifact
        OPTIONAL MATCH (source_artifact)-[rrel1:IS_RELATED]-(rel_node:Relationship)-[rrel2:IS_RELATED]-(definition_node:Node)
        WHERE rel_node.name = $definition_rel_identifier
        AND all(r IN [rrel1, rrel2] WHERE ( %(source_branch_filter)s ))
        RETURN
            definition_node,
            (rrel1.status = "active" AND rrel2.status = "active") AS definition_is_active,
            $source_branch_name IN [rrel1.branch, rrel2.branch] AS definition_on_source_branch
        ORDER BY rrel1.branch_level DESC, rrel1.branch_level DESC, rrel1.from DESC, rrel2.from DESC
        LIMIT 1
    }
    // -----------------------
    // get the artifact's checksum
    // -----------------------
    CALL {
        WITH source_artifact
        OPTIONAL MATCH (source_artifact)-[attr_rel:HAS_ATTRIBUTE]->(attr:Attribute)-[value_rel:HAS_VALUE]->(attr_val:AttributeValue)
        WHERE attr.name = "checksum"
        AND all(r IN [attr_rel, value_rel] WHERE ( %(source_branch_filter)s ))
        RETURN
            attr_val.value AS checksum,
            (attr_rel.status = "active" AND value_rel.status = "active") AS checksum_is_active,
            $source_branch_name IN [attr_rel.branch, value_rel.branch] AS checksum_on_source_branch
        ORDER BY value_rel.branch_level DESC, attr_rel.branch_level DESC, value_rel.from DESC, attr_rel.from DESC
        LIMIT 1
    }
    // -----------------------
    // get the artifact's storage_id
    // -----------------------
    CALL {
        WITH source_artifact
        OPTIONAL MATCH (source_artifact)-[attr_rel:HAS_ATTRIBUTE]->(attr:Attribute)-[value_rel:HAS_VALUE]->(attr_val:AttributeValue)
        WHERE attr.name = "storage_id"
        AND all(r IN [attr_rel, value_rel] WHERE ( %(source_branch_filter)s ))
        RETURN
            attr_val.value AS storage_id,
            (attr_rel.status = "active" AND value_rel.status = "active") AS storage_id_is_active,
            $source_branch_name IN [attr_rel.branch, value_rel.branch] AS storage_id_on_source_branch
        ORDER BY value_rel.branch_level DESC, attr_rel.branch_level DESC, value_rel.from DESC, attr_rel.from DESC
        LIMIT 1
    }
    WITH target_node, target_is_active, target_on_source_branch,
        definition_node, definition_is_active, definition_on_source_branch,
        checksum, checksum_is_active, checksum_on_source_branch,
        storage_id, storage_id_is_active, storage_id_on_source_branch
    WHERE (target_is_active AND target_on_source_branch)
    OR (definition_is_active AND definition_on_source_branch)
    OR (checksum_is_active AND checksum_on_source_branch)
    OR (storage_id_is_active AND storage_id_on_source_branch)
    RETURN CASE
        WHEN target_is_active = TRUE THEN target_node
        ELSE NULL
    END AS target_node,
    CASE
        WHEN definition_is_active = TRUE THEN definition_node
        ELSE NULL
    END AS definition_node,
    CASE
        WHEN checksum_is_active = TRUE THEN checksum
        ELSE NULL
    END AS source_checksum,
    CASE
        WHEN storage_id_is_active = TRUE THEN storage_id
        ELSE NULL
    END AS source_storage_id
}
CALL {
    // -----------------------
    // get the corresponding artifact on the target branch, if it exists
    // -----------------------
    WITH target_node, definition_node
    CALL {
        WITH target_node, definition_node
        OPTIONAL MATCH path = (target_node)-[trel1:IS_RELATED]-(trel_node:Relationship)-[trel2:IS_RELATED]-
        (target_artifact:%(artifact_kind)s)-[drel1:IS_RELATED]-(drel_node:Relationship)-[drel2:IS_RELATED]-(definition_node)
        WHERE trel_node.name = $target_rel_identifier
        AND drel_node.name = $definition_rel_identifier
        AND all(
            r IN relationships(path)
            WHERE r.branch = $target_branch_name
        )
        RETURN
            target_artifact,
            (trel1.status = "active" AND trel2.status = "active" AND drel1.status = "active" AND drel1.status = "active") AS artifact_is_active
        ORDER BY trel1.from DESC, trel2.from DESC, drel1.from DESC, drel2.from DESC
        LIMIT 1
    }
    WITH CASE
        WHEN artifact_is_active = TRUE THEN target_artifact
        ELSE NULL
    END as target_artifact
    // -----------------------
    // get the artifact's checksum on target branch
    // -----------------------
    CALL {
        WITH target_artifact
        OPTIONAL MATCH (target_artifact)-[attr_rel:HAS_ATTRIBUTE]->(attr:Attribute)-[value_rel:HAS_VALUE]->(attr_val:AttributeValue)
        WHERE attr.name = "checksum"
        AND attr_rel.branch = $target_branch_name
        AND value_rel.branch = $target_branch_name
        RETURN attr_val.value AS checksum, (attr_rel.status = "active" AND value_rel.status = "active") AS checksum_is_active
        ORDER BY value_rel.from DESC, attr_rel.from DESC
        LIMIT 1
    }
    // -----------------------
    // get the artifact's storage_id on target branch
    // -----------------------
    CALL {
        WITH target_artifact
        OPTIONAL MATCH (target_artifact)-[attr_rel:HAS_ATTRIBUTE]->(attr:Attribute)-[value_rel:HAS_VALUE]->(attr_val:AttributeValue)
        WHERE attr.name = "storage_id"
        AND attr_rel.branch = $target_branch_name
        AND value_rel.branch = $target_branch_name
        RETURN attr_val.value AS storage_id, (attr_rel.status = "active" AND value_rel.status = "active") AS storage_id_is_active
        ORDER BY value_rel.from DESC, attr_rel.from DESC
        LIMIT 1
    }
    RETURN target_artifact,
    CASE
        WHEN checksum_is_active = TRUE THEN checksum
        ELSE NULL
    END AS target_checksum,
    CASE
        WHEN storage_id_is_active = TRUE THEN storage_id
        ELSE NULL
    END AS target_storage_id
}
        """ % {"artifact_kind": InfrahubKind.ARTIFACT, "source_branch_filter": source_branch_filter}
        self.return_labels = [
            "source_artifact",
            "target_node",
            "definition_node",
            "source_checksum",
            "source_storage_id",
            "target_checksum",
            "target_storage_id",
        ]
        self.add_to_query(query=query)



from rich.console import Console

from ..schema.repository import InfrahubRepositoryConfig


def list_jinja2_transforms(config: InfrahubRepositoryConfig) -> None:
    console = Console()
    console.print(f"Jinja2 transforms defined in repository: {len(config.jinja2_transforms)}")

    for transform in config.jinja2_transforms:
        console.print(f"{transform.name} ({transform.template_path})")



from typing import Any

from pydantic import Field

from infrahub.message_bus import InfrahubMessage

from .models import InfrahubBranchEvent


class CommitUpdatedEvent(InfrahubBranchEvent):
    """Event generated when the the commit within a repository has been updated."""

    commit: str = Field(..., description="The commit the repository was updated to")
    repository_id: str = Field(..., description="The ID of the repository")
    repository_name: str = Field(..., description="The name of the repository")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.repository.update_commit"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.repository.{self.repository_id}",
            "infrahub.branch.name": self.branch,
            "infrahub.repository.name": self.repository_name,
            "infrahub.repository.id": self.repository_id,
        }

    def get_payload(self) -> dict[str, Any]:
        return {
            "branch": self.branch,
            "commit": self.commit,
            "repository_id": self.repository_id,
            "repository_name": self.repository_name,
        }

    def get_messages(self) -> list[InfrahubMessage]:
        return []



from __future__ import annotations

import uuid
from typing import Any, Optional, TypeAlias, Union

from infrahub_sdk.utils import deep_merge_dict
from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.constants import RESTRICTED_NAMESPACES
from infrahub.core.models import HashableModel
from infrahub.exceptions import SchemaNotFoundError

from .attribute_schema import AttributeSchema
from .basenode_schema import AttributePathParsingError, BaseNodeSchema, SchemaAttributePath, SchemaAttributePathValue
from .definitions.core import core_models
from .definitions.internal import internal
from .dropdown import DropdownChoice
from .generic_schema import GenericSchema
from .node_schema import NodeSchema
from .profile_schema import ProfileSchema
from .relationship_schema import RelationshipSchema

MainSchemaTypes: TypeAlias = Union[NodeSchema, GenericSchema, ProfileSchema]


# -----------------------------------------------------
# Extensions
#  For the initial implementation its possible to add attribute and relationships on Node
#  Later on we'll consider adding support for other Node attributes like inherited_from etc ...
#  And we'll look into adding support for Generic as well
class BaseNodeExtensionSchema(HashableModel):
    kind: str
    attributes: list[AttributeSchema] = Field(default_factory=list)
    relationships: list[RelationshipSchema] = Field(default_factory=list)


class NodeExtensionSchema(BaseNodeExtensionSchema):
    pass


class SchemaExtension(HashableModel):
    nodes: list[NodeExtensionSchema] = Field(default_factory=list)


class SchemaRoot(BaseModel):
    model_config = ConfigDict(extra="forbid")
    version: Optional[str] = Field(default=None)
    generics: list[GenericSchema] = Field(default_factory=list)
    nodes: list[NodeSchema] = Field(default_factory=list)
    extensions: SchemaExtension = SchemaExtension()

    @classmethod
    def has_schema(cls, values: dict[str, Any], name: str) -> bool:
        """Check if a schema exist locally as a node or as a generic."""

        available_schemas = [item.kind for item in values.get("nodes", []) + values.get("generics", [])]
        if name not in available_schemas:
            return False

        return True

    def get(self, name: str) -> Union[NodeSchema, GenericSchema]:
        """Check if a schema exist locally as a node or as a generic."""

        for item in self.nodes + self.generics:
            if item.kind == name:
                return item

        raise SchemaNotFoundError(branch_name="undefined", identifier=name)

    def validate_namespaces(self) -> list[str]:
        models = self.nodes + self.generics
        errors: list[str] = []
        for model in models:
            if model.namespace in RESTRICTED_NAMESPACES:
                errors.append(f"Restricted namespace '{model.namespace}' used on '{model.name}'")

        return errors

    def generate_uuid(self) -> None:
        """Generate UUID for all nodes, attributes & relationships
        Mainly useful during unit tests."""
        for node in self.nodes + self.generics:
            if not node.id:
                node.id = str(uuid.uuid4())
            for item in node.relationships + node.attributes:
                if not item.id:
                    item.id = str(uuid.uuid4())

    def merge(self, schema: SchemaRoot) -> SchemaRoot:
        """Return a new `SchemaRoot` after merging `self` with `schema`."""
        return SchemaRoot.model_validate(deep_merge_dict(dicta=self.model_dump(), dictb=schema.model_dump()))


internal_schema = internal.to_dict()

__all__ = [
    "AttributePathParsingError",
    "AttributeSchema",
    "BaseNodeSchema",
    "DropdownChoice",
    "GenericSchema",
    "MainSchemaTypes",
    "NodeSchema",
    "ProfileSchema",
    "RelationshipSchema",
    "SchemaAttributePath",
    "SchemaAttributePathValue",
    "SchemaRoot",
    "core_models",
    "internal_schema",
]



from typing import Optional

from infrahub_sdk.uuidt import UUIDT
from prefect import flow
from prefect.logging import get_run_logger

from infrahub.core.constants import InfrahubKind, ValidatorConclusion, ValidatorState
from infrahub.core.timestamp import Timestamp
from infrahub.message_bus import InfrahubMessage, Meta, messages
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices
from infrahub.workflows.utils import add_tags


@flow(
    name="artifact-definition-check",
    flow_run_name="Validating generation of artifacts for {message.artifact_definition.definition_name}",
)
async def check(message: messages.RequestArtifactDefinitionCheck, service: InfrahubServices) -> None:
    events: list[InfrahubMessage] = []
    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change], db_change=True)

    log = get_run_logger()
    artifact_definition = await service.client.get(
        kind=InfrahubKind.ARTIFACTDEFINITION,
        id=message.artifact_definition.definition_id,
        branch=message.source_branch,
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=message.proposed_change)

    validator_name = f"Artifact Validator: {message.artifact_definition.definition_name}"
    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []

    await proposed_change.validations.fetch()

    validator = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer
        if (
            existing_validator.typename == InfrahubKind.ARTIFACTVALIDATOR
            and existing_validator.definition.id == message.artifact_definition.definition_id
        ):
            validator = existing_validator

    if validator:
        validator.conclusion.value = ValidatorConclusion.UNKNOWN.value
        validator.state.value = ValidatorState.QUEUED.value
        validator.started_at.value = ""
        validator.completed_at.value = ""
        await validator.save()
    else:
        validator = await service.client.create(
            kind=InfrahubKind.ARTIFACTVALIDATOR,
            data={
                "label": validator_name,
                "proposed_change": message.proposed_change,
                "definition": message.artifact_definition.definition_id,
            },
        )
        await validator.save()

    await artifact_definition.targets.fetch()
    group = artifact_definition.targets.peer
    await group.members.fetch()

    existing_artifacts = await service.client.filters(
        kind=InfrahubKind.ARTIFACT,
        definition__ids=[message.artifact_definition.definition_id],
        include=["object"],
        branch=message.source_branch,
    )
    artifacts_by_member = {}
    for artifact in existing_artifacts:
        artifacts_by_member[artifact.object.peer.id] = artifact.id

    repository = message.branch_diff.get_repository(repository_id=message.artifact_definition.repository_id)
    requested_artifacts = 0
    impacted_artifacts = message.branch_diff.get_subscribers_ids(kind=InfrahubKind.ARTIFACT)
    for relationship in group.members.peers:
        member = relationship.peer
        artifact_id = artifacts_by_member.get(member.id)
        if _render_artifact(
            artifact_id=artifact_id,
            managed_branch=message.source_branch_sync_with_git,
            impacted_artifacts=impacted_artifacts,
        ):
            check_execution_id = str(UUIDT())
            check_execution_ids.append(check_execution_id)
            requested_artifacts += 1
            log.info(f"Trigger Artifact processing for {member.display_label}")
            events.append(
                messages.CheckArtifactCreate(
                    artifact_name=message.artifact_definition.artifact_name,
                    artifact_id=artifact_id,
                    artifact_definition=message.artifact_definition.definition_id,
                    commit=repository.source_commit,
                    content_type=message.artifact_definition.content_type,
                    transform_type=message.artifact_definition.transform_kind,
                    transform_location=message.artifact_definition.transform_location,
                    repository_id=repository.repository_id,
                    repository_name=repository.repository_name,
                    repository_kind=repository.kind,
                    branch_name=message.source_branch,
                    query=message.artifact_definition.query_name,
                    variables=member.extract(params=artifact_definition.parameters.value),
                    target_id=member.id,
                    target_name=member.display_label,
                    timeout=message.artifact_definition.timeout,
                    validator_id=validator.id,
                    meta=Meta(validator_execution_id=validator_execution_id, check_execution_id=check_execution_id),
                )
            )

    checks_in_execution = ",".join(check_execution_ids)
    await service.cache.set(
        key=f"validator_execution_id:{validator_execution_id}:checks",
        value=checks_in_execution,
        expires=KVTTL.TWO_HOURS,
    )
    events.append(
        messages.FinalizeValidatorExecution(
            start_time=Timestamp().to_string(),
            validator_id=validator.id,
            validator_execution_id=validator_execution_id,
            validator_type=InfrahubKind.ARTIFACTVALIDATOR,
        )
    )
    for event in events:
        event.assign_meta(parent=message)
        await service.send(message=event)


def _render_artifact(artifact_id: Optional[str], managed_branch: bool, impacted_artifacts: list[str]) -> bool:  # pylint: disable=unused-argument
    """Returns a boolean to indicate if an artifact should be generated or not.
    Will return true if:
        * The artifact_id wasn't set which could be that it's a new object that doesn't have a previous artifact
        * The source brance is not data only which would indicate that it could contain updates in git to the transform
        * The artifact_id exists in the impacted_artifacts list
    Will return false if:
        * The source branch is a data only branch and the artifact_id exists and is not in the impacted list
    """

    # if not artifact_id or managed_branch:
    #    return True
    # return artifact_id in impacted_artifacts
    # Temporary workaround tracked in https://github.com/opsmill/infrahub/issues/4991
    return True



from __future__ import annotations

import uuid
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import TYPE_CHECKING

import bcrypt
import jwt
from pydantic import BaseModel

from infrahub import config, models
from infrahub.core.account import validate_token
from infrahub.core.constants import AccountStatus, InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreAccount, CoreAccountGroup
from infrahub.core.registry import registry
from infrahub.exceptions import AuthorizationError, NodeNotFoundError

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreGenericAccount
    from infrahub.database import InfrahubDatabase


class AuthType(str, Enum):
    NONE = "none"
    JWT = "jwt"
    API = "api"


class AccountSession(BaseModel):
    authenticated: bool = True
    account_id: str
    session_id: str | None = None
    auth_type: AuthType

    @property
    def authenticated_by_jwt(self) -> bool:
        return self.auth_type == AuthType.JWT


async def validate_active_account(db: InfrahubDatabase, account_id: str) -> None:
    account: CoreGenericAccount = await NodeManager.get_one(db=db, id=account_id, raise_on_error=True)
    if account.status.value != AccountStatus.ACTIVE.value:
        raise AuthorizationError("This account has been deactivated")


async def authenticate_with_password(
    db: InfrahubDatabase, credentials: models.PasswordCredential, branch: str | None = None
) -> models.UserToken:
    selected_branch = await registry.get_branch(db=db, branch=branch)

    response: list[CoreGenericAccount] = await NodeManager.query(
        schema=InfrahubKind.GENERICACCOUNT,
        db=db,
        branch=selected_branch,
        filters={"name__value": credentials.username},
        limit=1,
    )

    if not response:
        raise NodeNotFoundError(
            branch_name=selected_branch.name,
            node_type=InfrahubKind.GENERICACCOUNT,
            identifier=credentials.username,
            message="That login user doesn't exist in the system",
        )

    account = response[0]
    if account.status.value != AccountStatus.ACTIVE.value:
        raise AuthorizationError("This account is not allowed to login")

    password = account.password.value
    valid_credentials = bcrypt.checkpw(credentials.password.encode("UTF-8"), str(password or "").encode("UTF-8"))
    if not valid_credentials:
        raise AuthorizationError("Incorrect password")

    now = datetime.now(tz=timezone.utc)
    refresh_expires = now + timedelta(seconds=config.SETTINGS.security.refresh_token_lifetime)

    session_id = await create_db_refresh_token(db=db, account_id=account.id, expiration=refresh_expires)
    access_token = generate_access_token(account_id=account.id, session_id=session_id)
    refresh_token = generate_refresh_token(account_id=account.id, session_id=session_id, expiration=refresh_expires)

    return models.UserToken(access_token=access_token, refresh_token=refresh_token)


async def create_db_refresh_token(db: InfrahubDatabase, account_id: str, expiration: datetime) -> uuid.UUID:
    obj = await Node.init(db=db, schema=InfrahubKind.REFRESHTOKEN)
    await obj.new(db=db, account=account_id, expiration=expiration.isoformat())
    await obj.save(db=db)
    return uuid.UUID(obj.id)


async def create_fresh_access_token(
    db: InfrahubDatabase, refresh_data: models.RefreshTokenData
) -> models.AccessTokenResponse:
    selected_branch = await registry.get_branch(db=db)

    refresh_token = await NodeManager.get_one(id=str(refresh_data.session_id), db=db)
    if not refresh_token:
        raise AuthorizationError("The provided refresh token has been invalidated in the database")

    account: CoreGenericAccount | None = await NodeManager.get_one(id=refresh_data.account_id, db=db)
    if not account:
        raise NodeNotFoundError(
            branch_name=selected_branch.name,
            node_type="Account",
            identifier=refresh_data.account_id,
            message="That login user doesn't exist in the system",
        )

    access_token = generate_access_token(account_id=account.id, session_id=refresh_data.session_id)

    return models.AccessTokenResponse(access_token=access_token)


async def signin_sso_account(db: InfrahubDatabase, account_name: str, sso_groups: list[str]) -> models.UserToken:
    account = await NodeManager.get_one_by_default_filter(db=db, id=account_name, kind=InfrahubKind.ACCOUNT)

    if not account:
        account = await Node.init(db=db, schema=InfrahubKind.ACCOUNT)
        await account.new(db=db, name=account_name, account_type="User", password=str(uuid.uuid4()))
        await account.save(db=db)

    if sso_groups:
        infrahub_groups = await NodeManager.query(
            db=db,
            schema=CoreAccountGroup,
            filters={"name__values": sso_groups},
            prefetch_relationships=True,
        )
        for group in infrahub_groups:
            members = await group.members.get_peers(db=db, branch_agnostic=True, peer_type=CoreAccount)
            if account.id not in members:
                await group.members.add(db=db, data=account)
                await group.members.save(db=db)

    now = datetime.now(tz=timezone.utc)
    refresh_expires = now + timedelta(seconds=config.SETTINGS.security.refresh_token_lifetime)
    session_id = await create_db_refresh_token(db=db, account_id=account.id, expiration=refresh_expires)
    access_token = generate_access_token(account_id=account.id, session_id=session_id)
    refresh_token = generate_refresh_token(account_id=account.id, session_id=session_id, expiration=refresh_expires)
    return models.UserToken(access_token=access_token, refresh_token=refresh_token)


def generate_access_token(account_id: str, session_id: uuid.UUID) -> str:
    now = datetime.now(tz=timezone.utc)

    access_expires = now + timedelta(seconds=config.SETTINGS.security.access_token_lifetime)
    access_data = {
        "sub": account_id,
        "iat": now,
        "nbf": now,
        "exp": access_expires,
        "fresh": False,
        "type": "access",
        "session_id": str(session_id),
    }
    access_token = jwt.encode(access_data, config.SETTINGS.security.secret_key, algorithm="HS256")
    return access_token


def generate_refresh_token(account_id: str, session_id: uuid.UUID, expiration: datetime) -> str:
    now = datetime.now(tz=timezone.utc)

    refresh_data = {
        "sub": account_id,
        "iat": now,
        "nbf": now,
        "exp": expiration,
        "fresh": False,
        "type": "refresh",
        "session_id": str(session_id),
    }
    refresh_token = jwt.encode(refresh_data, config.SETTINGS.security.secret_key, algorithm="HS256")
    return refresh_token


async def authentication_token(
    db: InfrahubDatabase, jwt_token: str | None = None, api_key: str | None = None
) -> AccountSession:
    if api_key:
        return await validate_api_key(db=db, token=api_key)
    if jwt_token:
        return await validate_jwt_access_token(token=jwt_token)

    return AccountSession(authenticated=False, account_id="anonymous", auth_type=AuthType.NONE)


async def validate_jwt_access_token(token: str) -> AccountSession:
    try:
        payload = jwt.decode(token, config.SETTINGS.security.secret_key, algorithms=["HS256"])
        account_id = payload["sub"]
        session_id = payload["session_id"]
    except jwt.ExpiredSignatureError:
        raise AuthorizationError("Expired Signature") from None
    except Exception:
        raise AuthorizationError("Invalid token") from None

    if payload["type"] == "access":
        return AccountSession(account_id=account_id, session_id=session_id, auth_type=AuthType.JWT)

    raise AuthorizationError("Invalid token, current token is not an access token")


async def validate_jwt_refresh_token(db: InfrahubDatabase, token: str) -> models.RefreshTokenData:
    try:
        payload = jwt.decode(token, config.SETTINGS.security.secret_key, algorithms=["HS256"])
        account_id = payload["sub"]
        session_id = payload["session_id"]
    except jwt.ExpiredSignatureError:
        raise AuthorizationError("Expired Signature") from None
    except Exception:
        raise AuthorizationError("Invalid token") from None

    await validate_active_account(db=db, account_id=str(account_id))

    if payload["type"] == "refresh":
        return models.RefreshTokenData(account_id=account_id, session_id=session_id)

    raise AuthorizationError("Invalid token, current token is not a refresh token")


async def validate_api_key(db: InfrahubDatabase, token: str) -> AccountSession:
    account_id = await validate_token(token=token, db=db)
    if not account_id:
        raise AuthorizationError("Invalid token")

    await validate_active_account(db=db, account_id=str(account_id))

    return AccountSession(account_id=account_id, auth_type=AuthType.API)


async def invalidate_refresh_token(db: InfrahubDatabase, token_id: str) -> None:
    refresh_token = await NodeManager.get_one(id=token_id, db=db)
    if refresh_token:
        await refresh_token.delete(db=db)



from pydantic import ConfigDict, Field

from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import ProposedChangeBranchDiff


class CheckRepositoryCheckDefinition(InfrahubMessage):
    """Triggers user defined checks to run based on a Check Definition."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    check_definition_id: str = Field(..., description="The unique ID of the check definition")
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    file_path: str = Field(..., description="The path and filename of the check")
    class_name: str = Field(..., description="The name of the class containing the check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Awaitable, Callable, MutableMapping, Optional, TypeVar

import aio_pika
import opentelemetry.instrumentation.aio_pika.span_builder
import ujson
from infrahub_sdk.uuidt import UUIDT
from opentelemetry.instrumentation.aio_pika import AioPikaInstrumentor
from opentelemetry.semconv.trace import SpanAttributes

from infrahub import config
from infrahub.components import ComponentType
from infrahub.log import clear_log_context, get_log_data
from infrahub.message_bus import InfrahubMessage, Meta, messages
from infrahub.message_bus.operations import execute_message
from infrahub.message_bus.types import MessageTTL
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from aio_pika.abc import (
        AbstractChannel,
        AbstractExchange,
        AbstractIncomingMessage,
        AbstractQueue,
        AbstractRobustConnection,
    )
    from opentelemetry.instrumentation.aio_pika.span_builder import SpanBuilder

    from infrahub.config import BrokerSettings
    from infrahub.services import InfrahubServices

MessageFunction = Callable[[InfrahubMessage], Awaitable[None]]
ResponseClass = TypeVar("ResponseClass")


AioPikaInstrumentor().instrument()


# TODO: remove this once https://github.com/open-telemetry/opentelemetry-python-contrib/issues/1835 is resolved
def patch_spanbuilder_set_channel() -> None:
    """
    The default SpanBuilder.set_channel does not work with aio_pika 9.1 and the refactored connection
    attribute
    """

    def set_channel(self: SpanBuilder, channel: AbstractChannel) -> None:
        if hasattr(channel, "_connection"):
            url = channel._connection.url
            self._attributes.update(
                {
                    SpanAttributes.NET_PEER_NAME: url.host,
                    SpanAttributes.NET_PEER_PORT: url.port,
                }
            )

    opentelemetry.instrumentation.aio_pika.span_builder.SpanBuilder.set_channel = set_channel  # type: ignore


async def _add_request_id(message: InfrahubMessage) -> None:
    log_data = get_log_data()
    message.meta.request_id = log_data.get("request_id", "")


class RabbitMQMessageBus(InfrahubMessageBus):
    def __init__(self, settings: Optional[BrokerSettings] = None) -> None:
        self.settings = settings or config.SETTINGS.broker
        self.channel: AbstractChannel
        self.exchange: AbstractExchange
        self.delayed_exchange: AbstractExchange
        self.service: InfrahubServices
        self.connection: AbstractRobustConnection
        self.callback_queue: AbstractQueue
        self.events_queue: AbstractQueue
        self.dlx: AbstractExchange
        self.message_enrichers: list[MessageFunction] = []

        self.loop = asyncio.get_running_loop()
        self.futures: MutableMapping[str, asyncio.Future] = {}

    async def initialize(self, service: InfrahubServices) -> None:
        patch_spanbuilder_set_channel()

        self.service = service
        self.connection = await aio_pika.connect_robust(
            host=self.settings.address,
            login=self.settings.username,
            password=self.settings.password,
            port=self.settings.service_port,
            virtualhost=self.settings.virtualhost,
            ssl=self.settings.tls_enabled,
            ssl_options={
                "no_verify_ssl": int(self.settings.tls_insecure),
                "cafile": self.settings.tls_ca_file or "",
            },
        )
        self.connection.reconnect_callbacks.add(self.on_reconnect, weak=False)

        await self._initialize_connection()

    async def _initialize_connection(self) -> None:
        self.channel = await self.connection.channel()

        if self.service.component_type == ComponentType.API_SERVER:
            await self._initialize_api_server()
        elif self.service.component_type == ComponentType.GIT_AGENT:
            await self._initialize_git_worker()

    async def shutdown(self) -> None:
        await self.connection.close()

    async def on_callback(self, message: AbstractIncomingMessage) -> None:
        if message.correlation_id:
            future: asyncio.Future = self.futures.pop(message.correlation_id)

            if future:
                future.set_result(message)
                return

        clear_log_context()
        if message.routing_key in messages.MESSAGE_MAP:
            await execute_message(routing_key=message.routing_key, message_body=message.body, service=self.service)
        else:
            self.service.log.error("Invalid message received", message=f"{message!r}")

    async def on_message(self, message: AbstractIncomingMessage) -> None:
        async with message.process():
            clear_log_context()
            if message.routing_key in messages.MESSAGE_MAP:
                await execute_message(routing_key=message.routing_key, message_body=message.body, service=self.service)
            else:
                self.service.log.error("Invalid message received", message=f"{message!r}")

    async def on_reconnect(
        self,
        weak: bool = False,  # pylint: disable=unused-argument
    ) -> None:
        self.service.log.info("Reconnected to RabbitMQ, reinitializing connection")
        await self._initialize_connection()

    async def _initialize_api_server(self) -> None:
        self.callback_queue = await self.channel.declare_queue(name=f"api-callback-{WORKER_IDENTITY}", exclusive=True)
        self.events_queue = await self.channel.declare_queue(name=f"api-events-{WORKER_IDENTITY}", exclusive=True)

        await self.callback_queue.consume(self.on_callback, no_ack=True)
        await self.events_queue.consume(self.on_callback, no_ack=True)
        self.exchange = await self.channel.declare_exchange(
            f"{self.settings.namespace}.events", type="topic", durable=True
        )
        self.dlx = await self.channel.declare_exchange(f"{self.settings.namespace}.dlx", type="topic", durable=True)

        queue = await self.channel.declare_queue(
            f"{self.settings.namespace}.rpcs",
            durable=True,
            arguments={"x-max-priority": 5},
        )

        self.delayed_exchange = await self.channel.declare_exchange(
            f"{self.settings.namespace}.delayed", type="headers", durable=True
        )
        for routing_key in self.worker_bindings:
            await queue.bind(self.exchange, routing_key=routing_key)
            await queue.bind(self.dlx, routing_key=routing_key)

        for ttl in MessageTTL.variations():
            ttl_queue = await self.channel.declare_queue(
                f"{self.settings.namespace}.delay.{ttl.name.lower()}_seconds",
                durable=True,
                arguments={
                    "x-dead-letter-exchange": self.dlx.name,
                    "x-message-ttl": ttl.value,
                    "x-max-priority": 5,
                },
            )
            await ttl_queue.bind(
                self.delayed_exchange,
                arguments={"x-match": "all", "delay": ttl.value},
            )

        for routing_key in self.event_bindings:
            await self.events_queue.bind(self.exchange, routing_key=routing_key)

        self.message_enrichers.append(_add_request_id)

    async def _initialize_git_worker(self) -> None:
        bindings = self.event_bindings + self.broadcasted_event_bindings
        events_queue = await self.channel.declare_queue(name=f"worker-events-{WORKER_IDENTITY}", exclusive=True)

        self.exchange = await self.channel.declare_exchange(
            f"{self.settings.namespace}.events", type="topic", durable=True
        )

        for routing_key in bindings:
            await events_queue.bind(self.exchange, routing_key=routing_key)
        self.delayed_exchange = await self.channel.get_exchange(name=f"{self.settings.namespace}.delayed")

        await events_queue.consume(callback=self.on_callback, no_ack=True)
        self.callback_queue = await self.channel.declare_queue(
            name=f"worker-callback-{WORKER_IDENTITY}", exclusive=True
        )
        await self.callback_queue.consume(self.on_callback, no_ack=True)

        message_channel = await self.connection.channel()
        await message_channel.set_qos(prefetch_count=self.settings.maximum_concurrent_messages)

        queue = await message_channel.get_queue(f"{self.settings.namespace}.rpcs")
        await queue.consume(callback=self.on_message, no_ack=False)

    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: Optional[MessageTTL] = None, is_retry: bool = False
    ) -> None:
        for enricher in self.message_enrichers:
            await enricher(message)
        message.assign_priority(priority=messages.message_priority(routing_key=routing_key))
        if delay:
            message.assign_header(key="delay", value=delay.value)
            await self.delayed_exchange.publish(self.format_message(message=message), routing_key=routing_key)
        else:
            await self.exchange.publish(self.format_message(message=message), routing_key=routing_key)

    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        await self.channel.default_exchange.publish(self.format_message(message=message), routing_key=routing_key)

    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        correlation_id = str(UUIDT())

        future = self.loop.create_future()
        self.futures[correlation_id] = future

        log_data = get_log_data()
        request_id = log_data.get("request_id", "")
        message.meta = Meta(request_id=request_id, correlation_id=correlation_id, reply_to=self.callback_queue.name)

        await self.service.send(message=message)

        response: AbstractIncomingMessage = await future
        data = ujson.loads(response.body)
        return response_class(**data)

    @staticmethod
    def format_message(message: InfrahubMessage) -> aio_pika.Message:
        pika_message = aio_pika.Message(
            body=message.body,
            content_type="application/json",
            content_encoding="utf-8",
            correlation_id=message.meta.correlation_id,
            reply_to=message.meta.reply_to,
            priority=message.meta.priority,
            headers=message.meta.headers,
            expiration=message.meta.expiration,
        )
        return pika_message



from __future__ import annotations

import ipaddress
import re
from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Union

import netaddr
import ujson
from infrahub_sdk.timestamp import TimestampFormatError
from infrahub_sdk.utils import is_valid_url
from infrahub_sdk.uuidt import UUIDT
from pydantic import BaseModel, Field

from infrahub import config
from infrahub.core import registry
from infrahub.core.constants import NULL_VALUE, AttributeDBNodeType, BranchSupportType, RelationshipStatus
from infrahub.core.property import FlagPropertyMixin, NodePropertyData, NodePropertyMixin
from infrahub.core.query.attribute import (
    AttributeGetQuery,
    AttributeUpdateFlagQuery,
    AttributeUpdateNodePropertyQuery,
    AttributeUpdateValueQuery,
)
from infrahub.core.query.node import AttributeFromDB, NodeListGetAttributeQuery
from infrahub.core.timestamp import Timestamp
from infrahub.core.utils import add_relationship, convert_ip_to_binary_str, update_relationships_to
from infrahub.exceptions import ValidationError
from infrahub.helpers import hash_password

from ..types import ATTRIBUTE_TYPES, LARGE_ATTRIBUTE_TYPES
from .constants.relationship_label import RELATIONSHIP_TO_NODE_LABEL, RELATIONSHIP_TO_VALUE_LABEL

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.schema import AttributeSchema
    from infrahub.database import InfrahubDatabase

# pylint: disable=redefined-builtin,c-extension-no-member,too-many-lines,too-many-public-methods


# Use a more user-friendly threshold than Neo4j one (8167 bytes).
MAX_STRING_LENGTH = 4096


def validate_string_length(value: Optional[str]) -> None:
    """
    Validates input string length does not exceed a given threshold, as Neo4J cannot index string values larger than 8167 bytes,
    see https://neo4j.com/developer/kb/index-limitations-and-workaround/.
    Note `value` parameter is optional as this function could be called from an attribute class
    with optional value such as StringOptional.
    """
    if value is None:
        return

    if 3 + len(value.encode("utf-8")) >= MAX_STRING_LENGTH:
        raise ValidationError(f"Text attribute length should be less than {MAX_STRING_LENGTH} characters.")


class AttributeCreateData(BaseModel):
    uuid: str
    name: str
    type: str
    branch: str
    branch_level: int
    branch_support: str
    status: str
    content: dict[str, Any]
    is_default: bool
    is_protected: bool
    is_visible: bool
    source_prop: list[NodePropertyData] = Field(default_factory=list)
    owner_prop: list[NodePropertyData] = Field(default_factory=list)
    node_type: AttributeDBNodeType = AttributeDBNodeType.DEFAULT


class BaseAttribute(FlagPropertyMixin, NodePropertyMixin):
    type: Optional[Union[type, tuple[type]]] = None

    _rel_to_node_label: str = RELATIONSHIP_TO_NODE_LABEL
    _rel_to_value_label: str = RELATIONSHIP_TO_VALUE_LABEL

    def __init__(  # pylint: disable=too-many-branches
        self,
        name: str,
        schema: AttributeSchema,
        branch: Branch,
        at: Timestamp,
        node: Node,
        id: Optional[str] = None,
        db_id: Optional[str] = None,
        data: Optional[Union[dict, str, AttributeFromDB]] = None,
        updated_at: Optional[Union[Timestamp, str]] = None,
        is_default: bool = False,
        is_from_profile: bool = False,
        **kwargs,
    ):
        self.id = id
        self.db_id = db_id

        self.updated_at = updated_at
        self.name = name
        self.node = node
        self.schema = schema
        self.branch = branch
        self.at = at
        self.is_default = is_default
        self.is_from_profile = is_from_profile
        self.from_pool: Optional[dict] = None

        self._init_node_property_mixin(kwargs)
        self._init_flag_property_mixin(kwargs)

        self.value = None

        if isinstance(data, AttributeFromDB):
            self.load_from_db(data=data)

        elif isinstance(data, dict):
            self.value = data.get("value")
            self.from_pool = data.get("from_pool")

            if "is_default" in data:
                self.is_default = data.get("is_default")
            if "is_from_profile" in data:
                self.is_from_profile = data.get("is_from_profile")

            fields_to_extract_from_data = ["id"] + self._flag_properties + self._node_properties
            for field_name in fields_to_extract_from_data:
                setattr(self, field_name, data.get(field_name, None))

            if not self.updated_at and "updated_at" in data:
                self.updated_at = Timestamp(data.get("updated_at"))
        elif data is None:
            self.is_default = True
        else:
            self.value = data

        # Assign default values
        if self.value is None and self.schema.default_value is not None:
            self.value = self.schema.default_value
            self.is_default = True

        if self.value is not None:
            self.validate(value=self.value, name=self.name, schema=self.schema)

        if self.is_enum and self.value:
            self.value = self.schema.convert_value_to_enum(self.value)

        if self.is_protected is None:
            self.is_protected = False

        if self.is_visible is None:
            self.is_visible = True

    @property
    def is_enum(self) -> bool:
        return bool(self.schema.enum)

    def get_branch_based_on_support_type(self) -> Branch:
        """If the attribute is branch aware, return the Branch object associated with this attribute
        If the attribute is branch agnostic return the Global Branch

        Returns:
            Branch:
        """
        if self.schema.branch == BranchSupportType.AGNOSTIC:
            return registry.get_global_branch()
        return self.branch

    @classmethod
    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        registry.attribute[cls.__name__] = cls

    def get_kind(self) -> str:
        return self.schema.kind

    def get_value(self) -> Any:
        if isinstance(self.value, Enum):
            return self.value.value
        return self.value

    def set_default_value(self) -> None:
        self.value = self.schema.default_value
        if self.is_enum and self.value:
            self.value = self.schema.convert_value_to_enum(self.value)

    @staticmethod
    def get_allowed_property_in_path() -> list[str]:
        return ["value"]

    @classmethod
    def validate(cls, value: Any, name: str, schema: AttributeSchema) -> bool:
        if value is None and schema.optional is False:
            raise ValidationError({name: f"A value must be provided for {name}"})
        if value is None and schema.optional is True:
            return True

        cls.validate_format(value=value, name=name, schema=schema)
        cls.validate_content(value=value, name=name, schema=schema)

        return True

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the format of the attribute.

        Args:
            value (Any): value to validate
            name (str): name of the attribute to include in a potential error message
            schema (AttributeSchema): schema for this attribute

        Raises:
            ValidationError: Format of the attribute value is not valid
        """
        value_to_check = value
        if schema.enum and isinstance(value, Enum):
            value_to_check = value.value
        if not isinstance(value_to_check, cls.type):  # pylint: disable=isinstance-second-argument-not-valid-type
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"})

    @classmethod
    def validate_content(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the content of the attribute.

        Args:
            value (Any): value to validate
            name (str): name of the attribute to include in a potential error message
            schema (AttributeSchema): schema for this attribute

        Raises:
            ValidationError: Content of the attribute value is not valid
        """
        if schema.regex:
            if schema.kind == "List":
                validation_values = [str(entry) for entry in value]
            else:
                validation_values = [str(value)]

            for validation_value in validation_values:
                try:
                    is_valid = re.match(pattern=schema.regex, string=str(validation_value))
                except re.error as exc:
                    raise ValidationError(
                        {name: f"The regex defined in the schema is not valid ({schema.regex!r})"}
                    ) from exc

                if not is_valid:
                    raise ValidationError({name: f"{validation_value} must conform with the regex: {schema.regex!r}"})

        if schema.min_length:
            if len(value) < schema.min_length:
                raise ValidationError({name: f"{value} must have a minimum length of {schema.min_length!r}"})

        if schema.max_length:
            if len(value) > schema.max_length:
                raise ValidationError({name: f"{value} must have a maximum length of {schema.max_length!r}"})

        if schema.enum:
            try:
                schema.convert_value_to_enum(value)
            except ValueError as exc:
                raise ValidationError({name: f"{value} must be one of {schema.enum!r}"}) from exc

    @classmethod
    def deserialize_from_string(cls, value_as_string: str) -> Any:
        """Return a value corresponding to the attribute type given it formatted as a string."""
        return cls.type(value_as_string)  # pylint: disable=not-callable

    def to_db(self) -> dict[str, Any]:
        """Return the properties of the AttributeValue node in Dict format."""
        data: dict[str, Any] = {"is_default": self.is_default}
        if self.value is None:
            data["value"] = NULL_VALUE
        else:
            serialized_value = self.serialize_value()
            if isinstance(serialized_value, str) and ATTRIBUTE_TYPES[self.schema.kind] not in LARGE_ATTRIBUTE_TYPES:
                # Perform validation here to avoid an extra serialization during validation step.
                # Standard non-str attributes (integer, boolean) do not exceed limit size related to neo4j indexing.
                validate_string_length(serialized_value)
            data["value"] = serialized_value

        return data

    def load_from_db(self, data: AttributeFromDB) -> None:
        self.value = self.value_from_db(data=data)
        self.is_default = data.is_default
        self.is_from_profile = data.is_from_profile

        self.id = data.attr_uuid
        self.db_id = data.attr_id

        for prop_name in self._flag_properties:
            if prop_name in data.flag_properties:
                setattr(self, prop_name, data.flag_properties[prop_name])

        for prop_name in self._node_properties:
            if prop_name in data.node_properties:
                setattr(self, prop_name, data.node_properties[prop_name].uuid)

        if not self.updated_at and data.updated_at:
            self.updated_at = Timestamp(data.updated_at)

    def value_from_db(self, data: AttributeFromDB) -> Any:
        if data.value == NULL_VALUE:
            return None
        return self.deserialize_value(data=data)

    def serialize_value(self) -> Any:
        """Serialize the value before storing it in the database."""
        if isinstance(self.value, Enum):
            return self.value.value
        return self.value

    def deserialize_value(self, data: AttributeFromDB) -> Any:
        """Deserialize the value coming from the database."""
        return data.value

    async def save(self, db: InfrahubDatabase, at: Optional[Timestamp] = None) -> bool:
        """Create or Update the Attribute in the database."""

        save_at = Timestamp(at)

        if not self.id or self.is_from_profile:
            return False

        return await self._update(at=save_at, db=db)

    async def delete(self, db: InfrahubDatabase, at: Optional[Timestamp] = None) -> bool:
        if not self.db_id:
            return False

        delete_at = Timestamp(at)

        query = await AttributeGetQuery.init(db=db, attr=self)
        await query.execute(db=db)
        results = query.get_results()

        if not results:
            return False

        properties_to_delete = []
        branch = self.get_branch_based_on_support_type()

        # Check all the relationship and update the one that are in the same branch
        rel_ids_to_update = set()
        for result in results:
            properties_to_delete.append((result.get_rel("r2").type, result.get_node("ap").element_id))

            await add_relationship(
                src_node_id=self.db_id,
                dst_node_id=result.get_node("ap").element_id,
                rel_type=result.get_rel("r2").type,
                branch_name=branch.name,
                branch_level=branch.hierarchy_level,
                at=delete_at,
                status=RelationshipStatus.DELETED,
                db=db,
            )

            for rel in result.get_rels():
                if rel.get("branch") == branch.name:
                    rel_ids_to_update.add(rel.element_id)

        if rel_ids_to_update:
            await update_relationships_to(ids=list(rel_ids_to_update), to=delete_at, db=db)

        await add_relationship(
            src_node_id=self.node.db_id,
            dst_node_id=self.db_id,
            rel_type="HAS_ATTRIBUTE",
            branch_name=branch.name,
            branch_level=branch.hierarchy_level,
            at=delete_at,
            status=RelationshipStatus.DELETED,
            db=db,
        )

        return True

    async def _update(self, db: InfrahubDatabase, at: Optional[Timestamp] = None) -> bool:
        """Update the attribute in the database.

        Get the current value
         - If the value is the same, do nothing
         - If the value is inherited and is different, raise error (for now just ignore)
         - If the value is different, create new node and update relationship

        """

        update_at = Timestamp(at)

        # Validate if the value is still correct, will raise a ValidationError if not
        self.validate(value=self.value, name=self.name, schema=self.schema)

        # Check if the current value is still the default one
        if self.is_default:
            if isinstance(self.value, Enum):
                has_default_value = self.schema.default_value == self.value.value
            else:
                has_default_value = self.schema.default_value == self.value
            if (self.schema.default_value is not None and not has_default_value) or (
                self.schema.default_value is None and self.value is not None
            ):
                self.is_default = False

        query = await NodeListGetAttributeQuery.init(
            db=db,
            ids=[self.node.id],
            fields={self.name: True},
            branch=self.branch,
            at=update_at,
            include_source=True,
            include_owner=True,
        )
        await query.execute(db=db)
        current_attr_data, current_attr_result = query.get_result_by_id_and_name(self.node.id, self.name)

        branch = self.get_branch_based_on_support_type()

        # ---------- Update the Value ----------
        if current_attr_data.content != self.to_db():
            # Create the new AttributeValue and update the existing relationship
            query = await AttributeUpdateValueQuery.init(db=db, attr=self, at=update_at)
            await query.execute(db=db)

            # TODO check that everything went well
            rel = current_attr_result.get_rel("r2")
            if rel.get("branch") == branch.name:
                await update_relationships_to([rel.element_id], to=update_at, db=db)

        # ---------- Update the Flags ----------
        SUPPORTED_FLAGS = (
            ("is_visible", "isv", "rel_isv"),
            ("is_protected", "isp", "rel_isp"),
        )

        for flag_name, _, rel_name in SUPPORTED_FLAGS:
            if current_attr_data.flag_properties[flag_name] != getattr(self, flag_name):
                query = await AttributeUpdateFlagQuery.init(db=db, attr=self, at=update_at, flag_name=flag_name)
                await query.execute(db=db)

                rel = current_attr_result.get(rel_name)
                if rel.get("branch") == branch.name:
                    await update_relationships_to([rel.element_id], to=update_at, db=db)

        # ---------- Update the Node Properties ----------
        for prop_name in self._node_properties:
            if getattr(self, f"{prop_name}_id") and not (
                prop_name in current_attr_data.node_properties
                and current_attr_data.node_properties[prop_name].uuid == getattr(self, f"{prop_name}_id")
            ):
                query = await AttributeUpdateNodePropertyQuery.init(
                    db=db, attr=self, at=update_at, prop_name=prop_name, prop_id=getattr(self, f"{prop_name}_id")
                )
                await query.execute(db=db)

                rel = current_attr_result.get(f"rel_{prop_name}")
                if rel and rel.get("branch") == branch.name:
                    await update_relationships_to([rel.element_id], to=update_at, db=db)

        return True

    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: Optional[dict] = None,
        related_node_ids: Optional[set] = None,
        filter_sensitive: bool = False,
        permissions: Optional[dict] = None,
        include_properties: bool = True,
    ) -> dict:
        """Generate GraphQL Payload for this attribute."""
        # pylint: disable=too-many-branches

        response: dict[str, Any] = {"id": self.id}

        if fields and isinstance(fields, dict):
            field_names = fields.keys()
        else:
            # REMOVED updated_at for now, need to investigate further how it's being used today
            field_names = ["__typename", "value"]
            if include_properties:
                field_names += self._node_properties + self._flag_properties

        for field_name in field_names:
            if field_name == "updated_at":
                if self.updated_at:
                    response[field_name] = await self.updated_at.to_graphql()
                else:
                    response[field_name] = None
                continue

            if field_name == "__typename":
                response[field_name] = self.get_kind()
                continue

            if field_name == "permissions":
                response[field_name] = {"update_value": permissions["update"]} if permissions else None
                continue

            if field_name in ["source", "owner"]:
                node_attr_getter = getattr(self, f"get_{field_name}")
                node_attr = await node_attr_getter(db=db)
                if not node_attr:
                    response[field_name] = None
                elif fields and isinstance(fields, dict):
                    response[field_name] = await node_attr.to_graphql(
                        db=db, fields=fields[field_name], related_node_ids=related_node_ids
                    )
                else:
                    response[field_name] = await node_attr.to_graphql(
                        db=db,
                        fields={"id": None, "display_label": None, "__typename": None},
                        related_node_ids=related_node_ids,
                    )
                continue

            if field_name.startswith("_"):
                field = getattr(self, field_name[1:])
            else:
                field = getattr(self, field_name)

            if field_name == "value" and isinstance(field, Enum):
                if config.SETTINGS.experimental_features.graphql_enums:
                    field = field.name
                else:
                    field = field.value
            if isinstance(field, str):
                response[field_name] = self._filter_sensitive(value=field, filter_sensitive=filter_sensitive)
            elif isinstance(field, (int, bool, dict, list)):
                response[field_name] = field

            if related_node_ids and self.is_from_profile and getattr(self, "source_id"):
                related_node_ids.add(getattr(self, "source_id"))

        return response

    def _filter_sensitive(self, value: str, filter_sensitive: bool) -> str:
        if filter_sensitive and self.schema.kind in ["HashedPassword", "Password"]:
            return "***"

        return value

    async def from_graphql(self, data: dict, db: InfrahubDatabase) -> bool:
        """Update attr from GraphQL payload"""

        changed = False
        if "value" in data:
            if self.is_enum:
                value_to_set = self.schema.convert_value_to_enum(data["value"])
            else:
                value_to_set = data["value"]
            if value_to_set != self.value:
                self.value = value_to_set
                changed = True
        elif "from_pool" in data:
            self.from_pool = data["from_pool"]
            await self.node.handle_pool(db=db, attribute=self, errors=[])
            changed = True

        if changed and self.is_from_profile:
            self.is_from_profile = False
            self.clear_source()

        if "is_default" in data and not self.is_default:
            self.is_default = True
            changed = True

            if "value" not in data:
                self.set_default_value()

        if "is_protected" in data and data["is_protected"] != self.is_protected:
            self.is_protected = data["is_protected"]
            changed = True
        if "is_visible" in data and data["is_visible"] != self.is_visible:
            self.is_visible = data["is_visible"]
            changed = True

        if "source" in data and data["source"] != self.source_id:
            self.source = data["source"]
            changed = True
        if "owner" in data and data["owner"] != self.owner_id:
            self.owner = data["owner"]
            changed = True

        return changed

    def get_db_node_type(self) -> AttributeDBNodeType:
        return AttributeDBNodeType.DEFAULT

    def get_create_data(self) -> AttributeCreateData:
        # pylint: disable=no-member
        branch = self.branch
        hierarchy_level = branch.hierarchy_level
        if self.schema.branch == BranchSupportType.AGNOSTIC:
            branch = registry.get_global_branch()
        elif self.schema.branch == BranchSupportType.LOCAL and self.node._schema.branch == BranchSupportType.AGNOSTIC:
            branch = registry.get_global_branch()
            hierarchy_level = 0
        data = AttributeCreateData(
            node_type=self.get_db_node_type(),
            uuid=str(UUIDT()),
            name=self.name,
            type=self.get_kind(),
            branch=branch.name,
            status="active",
            branch_level=hierarchy_level,
            branch_support=self.schema.branch.value,
            content=self.to_db(),
            is_default=self.is_default,
            is_protected=self.is_protected,
            is_visible=self.is_visible,
        )
        if self.source_id:
            data.source_prop.append(NodePropertyData(name="source", peer_id=self.source_id))

        if self.owner_id:
            data.owner_prop.append(NodePropertyData(name="owner", peer_id=self.owner_id))

        return data


class AnyAttribute(BaseAttribute):
    type = Any
    value: Any

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        pass

    @classmethod
    def deserialize_from_string(cls, value_as_string: str) -> Any:
        return value_as_string


class AnyAttributeOptional(AnyAttribute):
    @classmethod
    def deserialize_from_string(cls, value_as_string: str) -> Any:
        return value_as_string


class String(BaseAttribute):
    type = str
    value: str


class StringOptional(String):
    value: Optional[str]


class HashedPassword(BaseAttribute):
    type = str
    value: str

    def serialize_value(self) -> str:
        """Serialize the value before storing it in the database."""
        return hash_password(self.value)


class HashedPasswordOptional(HashedPassword):
    value: Optional[str]


class Integer(BaseAttribute):
    type = int
    value: int
    from_pool: Optional[str] = None

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """
        Make sure boolean objects are not accepted as value. Need to override `validate_format`
        as `isinstance(True, int)` is True.
        """

        value_to_check = value
        if schema.enum and isinstance(value, Enum):
            value_to_check = value.value

        # Note that we might want to do this check directly in parent function.
        if value_to_check.__class__ != cls.type:
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"})


class IntegerOptional(Integer):
    value: Optional[int]


class Boolean(BaseAttribute):
    type = bool
    value: bool


class BooleanOptional(Boolean):
    value: Optional[bool]


class DateTime(BaseAttribute):
    type = str
    value: str

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        super().validate_format(value=value, name=name, schema=schema)

        if not value and schema.optional:
            return

        try:
            Timestamp(value)
        except TimestampFormatError as exc:
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"}) from exc


class DateTimeOptional(DateTime):
    value: Optional[str]


class Dropdown(BaseAttribute):
    type = str
    value: str

    @property
    def color(self) -> str:
        """Return the color for the current value"""
        if self.schema.choices:
            selected = [choice for choice in self.schema.choices if choice.name == self.value]
            if selected and selected[0].color:
                return selected[0].color

        return ""

    @property
    def description(self) -> str:
        """Return the description for the current value"""
        if self.schema.choices:
            selected = [choice for choice in self.schema.choices if choice.name == self.value]
            if selected and selected[0].description:
                return selected[0].description

        return ""

    @property
    def label(self) -> str:
        """Return the label for the current value"""
        if self.schema.choices:
            selected = [choice for choice in self.schema.choices if choice.name == self.value]
            if selected and selected[0].label:
                return selected[0].label

        return ""

    @classmethod
    def validate_content(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the content of the dropdown."""
        super().validate_content(value=value, name=name, schema=schema)
        values = [choice.name for choice in schema.choices]
        if value not in values:
            raise ValidationError({name: f"{value} must be one of {', '.join(sorted(values))!r}"})


class DropdownOptional(Dropdown):
    value: Optional[str]


class URL(BaseAttribute):
    type = str
    value: str

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        super().validate_format(value=value, name=name, schema=schema)

        if not is_valid_url(value):
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"})


class URLOptional(URL):
    value: Optional[str]


class IPNetwork(BaseAttribute):
    type = str
    value: str

    @staticmethod
    def get_allowed_property_in_path() -> list[str]:
        return ["value", "version", "binary_address", "prefixlen"]

    @property
    def obj(self) -> Union[ipaddress.IPv4Network, ipaddress.IPv6Network]:
        """Return an ipaddress interface object."""
        if not self.value:
            raise ValueError("value for IPNetwork must be defined")
        return ipaddress.ip_network(str(self.value))

    @property
    def broadcast_address(self) -> Optional[str]:
        """Return the broadcast address of the ip network."""
        if not self.value:
            return None
        return str(self.obj.broadcast_address)

    @property
    def hostmask(self) -> Optional[str]:
        """Return the hostmask of the ip network."""
        if not self.value:
            return None
        return str(self.obj.hostmask)

    @property
    def netmask(self) -> Optional[str]:
        """Return the netmask of the ip network."""
        if not self.value:
            return None
        return str(self.obj.netmask)

    @property
    def network_address(self) -> Optional[str]:
        """Return the netmask of the ip network."""
        if not self.value:
            return None
        return str(self.obj.network_address)

    @property
    def network_address_integer(self) -> int:
        """Return the network address of the ip network in integer format."""
        return int(self.obj.network_address)

    @property
    def network_address_binary(self) -> str:
        """Return the network address of the ip network in binary format."""
        return convert_ip_to_binary_str(obj=self.obj)

    @property
    def prefixlen(self) -> Optional[int]:
        """Return the prefix length the ip network."""
        if not self.value:
            return None
        return ipaddress.ip_network(str(self.value)).prefixlen

    @property
    def num_addresses(self) -> Optional[int]:
        """Return the number of possible addresses in the ip network."""
        if not self.value:
            return None
        return ipaddress.ip_network(str(self.value)).num_addresses

    @property
    def version(self) -> Optional[int]:
        """Return the IP version of the ip network."""
        if not self.value:
            return None
        return ipaddress.ip_network(str(self.value)).version

    @property
    def with_hostmask(self) -> Optional[str]:
        """Return the network ip and the associated hostmask of the ip network."""
        if not self.value:
            return None
        return ipaddress.ip_network(str(self.value)).with_hostmask

    @property
    def with_netmask(self) -> Optional[str]:
        """Return the network ip and the associated netmask of the ip network."""
        if not self.value:
            return None
        return ipaddress.ip_network(str(self.value)).with_netmask

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the format of the attribute.

        Args:
            value (Any): value to validate
            name (str): name of the attribute to include in a potential error message
            schema (AttributeSchema): schema for this attribute

        Raises:
            ValidationError: Format of the attribute value is not valid
        """
        super().validate_format(value=value, name=name, schema=schema)

        try:
            ipaddress.ip_network(value)
        except ValueError as exc:
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"}) from exc

    def serialize_value(self) -> str:
        """Serialize the value before storing it in the database. If network is an IPv6 network, it is converted to collapsed form."""

        return ipaddress.ip_network(self.value).with_prefixlen

    def get_db_node_type(self) -> AttributeDBNodeType:
        if self.value is not None:
            return AttributeDBNodeType.IPNETWORK
        return AttributeDBNodeType.DEFAULT

    def to_db(self) -> dict[str, Any]:
        data = super().to_db()

        if self.value is not None:
            data["version"] = self.version
            data["binary_address"] = self.network_address_binary
            data["prefixlen"] = self.prefixlen
            # data["num_addresses"] = self.num_addresses

        return data


class IPNetworkOptional(IPNetwork):
    value: Optional[str]


class IPHost(BaseAttribute):
    type = str
    value: str

    @staticmethod
    def get_allowed_property_in_path() -> list[str]:
        return ["value", "version", "binary_address"]

    @property
    def obj(self) -> Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface]:
        """Return the ip adress without a prefix or subnet mask."""
        if not self.value:
            raise ValueError("value for IPHost must be defined")
        return ipaddress.ip_interface(str(self.value))

    @property
    def ip(self) -> Optional[str]:
        """Return the ip adress without a prefix or subnet mask."""
        if not self.value:
            return None
        return str(self.obj.ip)

    @property
    def hostmask(self) -> Optional[str]:
        """Return the hostmask of the ip address."""
        if not self.value:
            return None
        return str(self.obj.hostmask)

    @property
    def netmask(self) -> Optional[str]:
        """Return the netmask of the ip address."""
        if not self.value:
            return None
        return str(self.obj.netmask)

    @property
    def network(self) -> Optional[str]:
        """Return the network encapsuling the ip address."""
        if not self.value:
            return None
        return str(self.obj.network)

    @property
    def prefixlen(self) -> Optional[int]:
        """Return the prefix length of the ip address."""
        if not self.value:
            return None
        return self.obj.network.prefixlen

    @property
    def version(self) -> Optional[int]:
        """Return the IP version of the ip address."""
        if not self.value:
            return None
        return self.obj.version

    @property
    def with_hostmask(self) -> Optional[str]:
        """Return the ip address and the associated hostmask of the ip address."""
        if not self.value:
            return None
        return self.obj.with_hostmask

    @property
    def with_netmask(self) -> Optional[str]:
        """Return the ip address and the associated netmask of the ip address."""
        if not self.value:
            return None
        return self.obj.with_netmask

    @property
    def ip_integer(self) -> int:
        """Return the ip address in binary format."""
        return int(self.obj)

    @property
    def ip_binary(self) -> str:
        """Return the ip address in binary format."""
        return convert_ip_to_binary_str(obj=self.obj)

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the format of the attribute.

        Args:
            value (Any): value to validate
            name (str): name of the attribute to include in a potential error message
            schema (AttributeSchema): schema for this attribute

        Raises:
            ValidationError: Format of the attribute value is not valid
        """
        super().validate_format(value=value, name=name, schema=schema)

        try:
            ipaddress.ip_interface(value)
        except ValueError as exc:
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"}) from exc

    def serialize_value(self) -> str:
        """Adds a prefix to address before storing it in the database. If address in an IPv6 address, it is converted to collapsed form."""

        return ipaddress.ip_interface(self.value).with_prefixlen

    def get_db_node_type(self) -> AttributeDBNodeType:
        if self.value is not None:
            return AttributeDBNodeType.IPHOST
        return AttributeDBNodeType.DEFAULT

    def to_db(self) -> dict[str, Any]:
        data = super().to_db()

        if self.value is not None:
            data["version"] = self.version
            data["binary_address"] = self.ip_binary
            data["prefixlen"] = self.prefixlen

        return data


class IPHostOptional(IPHost):
    value: Optional[str]


class MacAddress(BaseAttribute):
    type = str
    value: str

    @property
    def obj(self) -> netaddr.EUI:
        """Return the MAC adress."""
        if not self.value:
            raise ValueError("value for MAC address must be defined")
        return netaddr.EUI(addr=self.value)

    @property
    def oui(self) -> Optional[str]:
        """Return the OUI (Organisationally Unique Identifier) for the MAC address."""
        if not self.value:
            return None
        try:
            return str(self.obj.oui)
        except netaddr.NotRegisteredError:
            # Workaround OUI lookup failure
            # netaddr might not be up-to-date, or actual OUI just not exist (i.e. random mac address)
            return str(self.obj).removesuffix(f"-{self.ei}")

    @property
    def ei(self) -> Optional[str]:
        """Return the EI (Extension Identifier) for the MAC address."""
        if not self.value:
            return None
        return self.obj.ei

    @property
    def version(self) -> Optional[int]:
        """Return the version of the MAC address."""
        if not self.value:
            return None
        return self.obj.version

    @property
    def binary(self) -> Optional[str]:
        """Return the MAC address in binary format."""
        if not self.value:
            return None
        return self.obj.bin

    @property
    def eui48(self) -> Optional[str]:
        if not self.value:
            return None
        return self.obj.format(dialect=netaddr.mac_eui48)

    @property
    def eui64(self) -> Optional[str]:
        if not self.value:
            return None
        return str(self.obj.eui64())

    @property
    def bare(self) -> Optional[str]:
        if not self.value:
            return None
        return self.obj.format(dialect=netaddr.mac_bare)

    @property
    def dot_notation(self) -> Optional[str]:
        if not self.value:
            return None
        return self.obj.format(dialect=netaddr.mac_cisco)

    @property
    def semicolon_notation(self) -> Optional[str]:
        if not self.value:
            return None
        return self.obj.format(dialect=netaddr.mac_unix)

    @property
    def split_notation(self) -> Optional[str]:
        if not self.value:
            return None
        return self.obj.format(dialect=netaddr.mac_pgsql)

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        """Validate the format of the attribute.

        Args:
            value (Any): value to validate
            name (str): name of the attribute to include in a potential error message
            schema (AttributeSchema): schema for this attribute

        Raises:
            ValidationError: Format of the attribute value is not valid
        """
        super().validate_format(value=value, name=name, schema=schema)

        if not netaddr.valid_mac(addr=str(value)):
            raise ValidationError({name: f"{value} is not a valid {schema.kind}"})

    def serialize_value(self) -> str:
        """Serialize the value as standard EUI-48 or EUI-64 before storing it in the database."""
        return str(netaddr.EUI(addr=self.value))


class MacAddressOptional(MacAddress):
    value: Optional[str]


class ListAttribute(BaseAttribute):
    type = list
    value: list[Any]

    @classmethod
    def deserialize_from_string(cls, value_as_string: str) -> Any:
        if value_as_string:
            return ujson.loads(value_as_string)
        return []

    def serialize_value(self) -> str:
        """Serialize the value before storing it in the database."""
        return ujson.dumps(self.value)

    def deserialize_value(self, data: AttributeFromDB) -> Any:
        """Deserialize the value (potentially) coming from the database."""
        if isinstance(data.value, (str, bytes)):
            return ujson.loads(data.value)
        return data.value

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        value_to_check = value

        if isinstance(value, str):
            try:
                value_to_check = cls.deserialize_from_string(value_as_string=value)
            except ujson.JSONDecodeError as exc:
                raise ValidationError({name: f"{value} is not a valid JSON list"}) from exc
        super().validate_format(value=value_to_check, name=name, schema=schema)


class ListAttributeOptional(ListAttribute):
    value: Optional[list[Any]]


class JSONAttribute(BaseAttribute):
    type = (dict, list)
    value: Union[dict, list]

    @classmethod
    def deserialize_from_string(cls, value_as_string: str) -> Any:
        if value_as_string:
            return ujson.loads(value_as_string)
        return {}

    def serialize_value(self) -> str:
        """Serialize the value before storing it in the database."""
        return ujson.dumps(self.value)

    def deserialize_value(self, data: AttributeFromDB) -> Any:
        """Deserialize the value (potentially) coming from the database."""
        if data.value and isinstance(data.value, (str, bytes)):
            return ujson.loads(data.value)
        return data.value

    @classmethod
    def validate_format(cls, value: Any, name: str, schema: AttributeSchema) -> None:
        value_to_check = value

        if isinstance(value, str):
            try:
                value_to_check = cls.deserialize_from_string(value_as_string=value)
            except ujson.JSONDecodeError as exc:
                raise ValidationError({name: f"{value} is not valid JSON"}) from exc
        super().validate_format(value=value_to_check, name=name, schema=schema)


class JSONAttributeOptional(JSONAttribute):
    value: Optional[Union[dict, list]]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration011Query01(Query):
    name = "migration_011_01"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        query = """
        // get all the SchemaRelationship nodes for 'profiles' relationships
        MATCH (sr_to_delete:SchemaRelationship)-[HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[HAS_VALUE]->(av:AttributeValue {value: "profiles"})
        // get the Relationship node for each 'profiles' SchemaRelationship
        MATCH (node_rel_to_delete:Relationship)-[IS_RELATED]-(sr_to_delete)
        WITH node_rel_to_delete, sr_to_delete
        // get all the Attributes and AttributeValues of the 'profiles' SchemaRelationships
        MATCH (sr_to_delete)-[HAS_ATTRIBUTE]->(attribute_to_delete:Attribute)-[HAS_VALUE]->(av_to_maybe_delete:AttributeValue)
        // detach delete the Relationship, SchemaRelationship, and Attribute nodes
        DETACH DELETE node_rel_to_delete, sr_to_delete, attribute_to_delete
        // delete the AttributeValue nodes from earlier that are not attached to anything
        WITH av_to_maybe_delete
        WHERE NOT EXISTS {
            MATCH (av_to_maybe_delete)-[]-()
        }
        DELETE av_to_maybe_delete
        """
        self.add_to_query(query)


class Migration011(GraphMigration):
    name: str = "011_remove_profiles_relationship_schemas"
    queries: Sequence[type[Query]] = [Migration011Query01]
    minimum_version: int = 10

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        return result



from typing import Union

from pydantic import BaseModel, Field

from infrahub.core.branch import Branch
from infrahub.core.path import SchemaPath
from infrahub.core.schema import GenericSchema, NodeSchema


class SchemaConstraintValidatorRequest(BaseModel):
    branch: Branch = Field(..., description="The name of the branch to target")
    constraint_name: str = Field(..., description="The name of the constraint to validate")
    node_schema: Union[NodeSchema, GenericSchema] = Field(..., description="Schema of Node or Generic to validate")
    schema_path: SchemaPath = Field(..., description="SchemaPath to the element of the schema to validate")


class SchemaViolation(BaseModel):
    node_id: str
    node_kind: str
    display_label: str
    full_display_label: str
    message: str = ""



from typing import Optional

from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.registry import registry
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase

from .interface import MutationNodeGetterInterface


class MutationNodeGetterByHfid(MutationNodeGetterInterface):
    def __init__(self, db: InfrahubDatabase, node_manager: NodeManager) -> None:
        self.db = db
        self.node_manager = node_manager

    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Optional[Node]:
        if not node_schema.human_friendly_id:
            return None

        if "hfid" in data:
            return await self.node_manager.get_one_by_hfid(
                db=self.db,
                hfid=data["hfid"],
                kind=node_schema.kind,
                branch=branch,
            )

        for component in node_schema.human_friendly_id:
            name = component.split("__")[0]
            if name not in data.keys():
                # The update neither includes "hfid" or all components to form an hfid:
                return None

        schema_branch = registry.schema.get_schema_branch(name=branch.name)
        hfid: list[str] = []
        for component in node_schema.human_friendly_id:
            attribute_path = node_schema.parse_schema_path(path=component, schema=schema_branch)

            if attribute_path.is_type_attribute and attribute_path.attribute_schema:
                hfid_component = data[attribute_path.attribute_schema.name].get(attribute_path.attribute_property_name)
                if hfid_component is not None:
                    hfid.append(str(hfid_component))
            if (
                attribute_path.relationship_schema
                and attribute_path.related_schema
                and attribute_path.attribute_property_name
                and attribute_path.attribute_schema
            ):
                related_node = await self.node_manager.find_object(
                    db=self.db,
                    kind=attribute_path.relationship_schema.peer,
                    branch=branch,
                    id=data[attribute_path.relationship_schema.name].get("id"),
                    hfid=data[attribute_path.relationship_schema.name].get("hfid"),
                )
                relationship_attribute = getattr(related_node, attribute_path.attribute_schema.name)
                relationship_attribute_value = getattr(relationship_attribute, attribute_path.attribute_property_name)
                if relationship_attribute_value is None:
                    return None

                hfid.append(str(relationship_attribute_value))

        if len(hfid) == len(node_schema.human_friendly_id):
            return await self.node_manager.get_one_by_hfid(
                db=self.db,
                hfid=hfid,
                kind=node_schema.kind,
                branch=branch,
            )

        return None



from typing import Optional

from .schema.attribute_name_update import AttributeNameUpdateMigration
from .schema.node_attribute_add import NodeAttributeAddMigration
from .schema.node_attribute_remove import NodeAttributeRemoveMigration
from .schema.node_kind_update import NodeKindUpdateMigration
from .schema.node_remove import NodeRemoveMigration
from .schema.placeholder_dummy import PlaceholderDummyMigration
from .shared import SchemaMigration

MIGRATION_MAP: dict[str, Optional[type[SchemaMigration]]] = {
    "node.remove": NodeRemoveMigration,
    "node.branch.update": None,
    "node.attribute.add": NodeAttributeAddMigration,
    "node.attribute.remove": NodeAttributeRemoveMigration,
    "node.name.update": NodeKindUpdateMigration,
    "node.namespace.update": NodeKindUpdateMigration,
    "node.relationship.remove": PlaceholderDummyMigration,
    "attribute.name.update": AttributeNameUpdateMigration,
    "attribute.branch.update": None,
    "relationship.branch.update": None,
    "relationship.direction.update": None,
    "relationship.identifier.update": PlaceholderDummyMigration,
    "relationship.hierarchical.update": None,
}



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from typing import Optional

from pydantic import Field

from infrahub.core.constants import (
    AllowOverrideType,
    BranchSupportType,
    HashableModelState,
    RelationshipCardinality,
    RelationshipDeleteBehavior,
    RelationshipDirection,
    RelationshipKind,
)  # noqa: TC001
from infrahub.core.models import HashableModel


class GeneratedRelationshipSchema(HashableModel):
    id: Optional[str] = Field(
        default=None, description="The ID of the relationship schema", json_schema_extra={"update": "not_applicable"}
    )
    name: str = Field(
        ...,
        description="Relationship name, must be unique within a model and must be all lowercase.",
        pattern=r"^[a-z0-9\_]+$",
        min_length=3,
        max_length=32,
        json_schema_extra={"update": "allowed"},
    )
    peer: str = Field(
        ...,
        description="Type (kind) of objects supported on the other end of the relationship.",
        pattern="^[A-Z][a-zA-Z0-9]+$",
        json_schema_extra={"update": "validate_constraint"},
    )
    kind: RelationshipKind = Field(
        default=RelationshipKind.GENERIC,
        description="Defines the type of the relationship.",
        json_schema_extra={"update": "allowed"},
    )
    label: Optional[str] = Field(
        default=None,
        description="Human friendly representation of the name. Will be autogenerated if not provided",
        max_length=32,
        json_schema_extra={"update": "allowed"},
    )
    description: Optional[str] = Field(
        default=None,
        description="Short description of the relationship.",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )
    identifier: Optional[str] = Field(
        default=None,
        description="Unique identifier of the relationship within a model, identifiers must match to traverse a relationship on both direction.",
        pattern=r"^[a-z0-9\_]+$",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )
    cardinality: RelationshipCardinality = Field(
        default=RelationshipCardinality.MANY,
        description="Defines how many objects are expected on the other side of the relationship.",
        json_schema_extra={"update": "validate_constraint"},
    )
    min_count: int = Field(
        default=0,
        description="Defines the minimum objects allowed on the other side of the relationship.",
        json_schema_extra={"update": "validate_constraint"},
    )
    max_count: int = Field(
        default=0,
        description="Defines the maximum objects allowed on the other side of the relationship.",
        json_schema_extra={"update": "validate_constraint"},
    )
    order_weight: Optional[int] = Field(
        default=None,
        description="Number used to order the relationship in the frontend (table and view). Lowest value will be ordered first.",
        json_schema_extra={"update": "allowed"},
    )
    optional: bool = Field(
        default=True,
        description="Indicate if this relationship is mandatory or optional.",
        json_schema_extra={"update": "validate_constraint"},
    )
    branch: Optional[BranchSupportType] = Field(
        default=None,
        description="Type of branch support for the relatioinship, if not defined it will be determine based both peers.",
        json_schema_extra={"update": "not_supported"},
    )
    inherited: bool = Field(
        default=False,
        description="Internal value to indicate if the relationship was inherited from a Generic node.",
        json_schema_extra={"update": "not_applicable"},
    )
    direction: RelationshipDirection = Field(
        default=RelationshipDirection.BIDIR,
        description="Defines the direction of the relationship,  Unidirectional relationship are required when the same model is on both side.",
        json_schema_extra={"update": "not_supported"},
    )
    hierarchical: Optional[str] = Field(
        default=None,
        description="Internal attribute to track the type of hierarchy this relationship is part of, must match a valid Generic Kind",
        json_schema_extra={"update": "not_supported"},
    )
    state: HashableModelState = Field(
        default=HashableModelState.PRESENT,
        description="Expected state of the relationship after loading the schema",
        json_schema_extra={"update": "not_applicable"},
    )
    on_delete: Optional[RelationshipDeleteBehavior] = Field(
        default=None,
        description="Default is no-action. If cascade, related node(s) are deleted when this node is deleted.",
        json_schema_extra={"update": "allowed"},
    )
    allow_override: AllowOverrideType = Field(
        default=AllowOverrideType.ANY,
        description="Type of allowed override for the relationship.",
        json_schema_extra={"update": "allowed"},
    )
    read_only: bool = Field(
        default=False,
        description="Set the relationship as read-only, users won't be able to change its value.",
        json_schema_extra={"update": "allowed"},
    )
    deprecation: Optional[str] = Field(
        default=None,
        description="Mark relationship as deprecated and provide a user-friendly message to display",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from typing_extensions import Self

from infrahub.core.constants import SchemaPathType
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.path import SchemaPath

from ..schema.node_attribute_add import NodeAttributeAddMigration
from ..shared import InternalSchemaMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration009(InternalSchemaMigration):
    name: str = "009_add_generate_profile_attr"
    minimum_version: int = 8

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        internal_schema = cls.get_internal_schema()
        schema_node = internal_schema.get_node(name="SchemaNode")

        migrations = [
            NodeAttributeAddMigration(
                new_node_schema=schema_node,
                previous_node_schema=schema_node,
                schema_path=SchemaPath(
                    schema_kind="SchemaNode", path_type=SchemaPathType.ATTRIBUTE, field_name="generate_profile"
                ),
            ),
        ]
        return cls(migrations=migrations, **kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()
        return result



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field
from typing_extensions import Self

from infrahub.core.account import GlobalPermission
from infrahub.core.node import Node
from infrahub.core.protocols import CoreMenuItem
from infrahub.core.schema import GenericSchema, MainSchemaTypes, NodeSchema, ProfileSchema

from .constants import MenuSection

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


def get_full_name(obj: CoreMenuItem | NodeSchema | GenericSchema | ProfileSchema) -> str:
    if isinstance(obj, (NodeSchema, GenericSchema, ProfileSchema)):
        return _get_full_name_schema(obj)
    return _get_full_name_node(obj)


def _get_full_name_node(obj: CoreMenuItem) -> str:
    return f"{obj.namespace.value}{obj.name.value}"


def _get_full_name_schema(node: MainSchemaTypes) -> str:
    return f"{node.namespace}{node.name}"


@dataclass
class MenuDict:
    data: dict[str, MenuItemDict] = field(default_factory=dict)

    def find_item(self, name: str) -> MenuItemDict | None:
        return self._find_child_item(name=name, children=self.data)

    @classmethod
    def _find_child_item(cls, name: str, children: dict[str, MenuItemDict]) -> MenuItemDict | None:
        if name in children.keys():
            return children[name]

        for child in children.values():
            if not child.children:
                continue
            found = cls._find_child_item(name=name, children=child.children)
            if found:
                return found
        return None

    def to_rest(self) -> Menu:
        data: dict[str, list[MenuItemList]] = {}

        for section in [MenuSection.INTERNAL, MenuSection.OBJECT]:
            item_per_section = [
                value.to_list() for value in self.data.values() if value.section == section and value.hidden is False
            ]
            data[section.value] = sorted(item_per_section, key=lambda d: d.order_weight)

        return Menu(sections=data)

    # @staticmethod
    # def _sort_menu_items(items: dict[str, MenuItem]) -> dict[str, MenuItem]:
    #     sorted_dict = dict(sorted(items.items(), key=lambda x: (x[1].order_weight, x[0]), reverse=False))
    #     return sorted_dict


@dataclass
class Menu:
    sections: dict[str, list[MenuItemList]] = field(default_factory=dict)


class MenuItem(BaseModel):
    identifier: str = Field(..., description="Unique identifier for this menu item")
    label: str = Field(..., description="Title of the menu item")
    path: str = Field(default="", description="URL endpoint if applicable")
    icon: str = Field(default="", description="The icon to show for the current view")
    kind: str = Field(default="", description="Kind of the model associated with this menuitem if applicable")
    order_weight: int = 5000
    section: MenuSection = MenuSection.OBJECT
    permissions: list[str] = Field(default_factory=list)

    @classmethod
    def from_node(cls, obj: CoreMenuItem) -> Self:
        return cls(
            identifier=get_full_name(obj),
            label=obj.label.value or "",
            icon=obj.icon.value or "",
            order_weight=obj.order_weight.value,
            path=obj.path.value or "",
            kind=obj.kind.value or "",
            section=obj.section.value,
            permissions=obj.required_permissions.value or [],
        )

    @classmethod
    def from_schema(cls, model: NodeSchema | GenericSchema | ProfileSchema) -> Self:
        return cls(
            identifier=get_full_name(model),
            label=model.label or model.kind,
            path=f"/objects/{model.kind}",
            icon=model.icon or "",
            kind=model.kind,
        )


class MenuItemDict(MenuItem):
    hidden: bool = False
    children: dict[str, MenuItemDict] = Field(default_factory=dict, description="Child objects")

    def to_list(self) -> MenuItemList:
        data = self.model_dump(exclude={"children"})
        unsorted_children = [child.to_list() for child in self.children.values() if child.hidden is False]
        data["children"] = sorted(unsorted_children, key=lambda d: d.order_weight)
        return MenuItemList(**data)

    def get_global_permissions(self) -> list[GlobalPermission]:
        permissions: list[GlobalPermission] = []
        for permission in self.permissions:
            if not permission.startswith("global"):
                continue
            permissions.append(GlobalPermission.from_string(input=permission))
        return permissions


class MenuItemList(MenuItem):
    children: list[MenuItemList] = Field(default_factory=list, description="Child objects")


class MenuItemDefinition(BaseModel):
    namespace: str
    name: str
    label: str
    description: str = ""
    icon: str = ""
    protected: bool = False
    path: str = ""
    kind: str = ""
    section: MenuSection = MenuSection.OBJECT
    order_weight: int = 2000
    permissions: list[str] = Field(default_factory=list)
    children: list[MenuItemDefinition] = Field(default_factory=list)

    async def to_node(self, db: InfrahubDatabase, parent: CoreMenuItem | None = None) -> CoreMenuItem:
        obj = await Node.init(db=db, schema=CoreMenuItem)
        await obj.new(
            db=db,
            namespace=self.namespace,
            name=self.name,
            label=self.label,
            path=self.get_path(),
            description=self.description or None,
            icon=self.icon or None,
            protected=self.protected,
            section=self.section.value,
            order_weight=self.order_weight,
            parent=parent.id if parent else None,
            required_permissions=self.permissions,
        )
        return obj

    def get_path(self) -> str | None:
        if self.path:
            return self.path

        if self.kind:
            return f"/objects/{self.kind}"

        return None

    @property
    def full_name(self) -> str:
        return f"{self.namespace}{self.name}"



from infrahub.message_bus import messages
from infrahub.services import InfrahubServices


async def mutated(
    message: messages.EventNodeMutated,
    service: InfrahubServices,
) -> None:
    """Event posted when a node is mutated"""
    # Note for now this is only kept to facilitate publishing to other queues in the future
    # This operation doesn't have a flow defined to avoid having the workers need to register
    # the results in Prefect when they don't actually do anything aside from add noise.
    service.log.debug(
        "Mutation on node",
        branch=message.branch,
        node_id=message.node_id,
        action=message.action,
        kind=message.kind,
        data=message.data,
    )



from typing import Any

from pydantic import Field

from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.messages.refresh_registry_branches import RefreshRegistryBranches

from .models import InfrahubBranchEvent


class SchemaUpdatedEvent(InfrahubBranchEvent):
    """Event generated when the schema within a branch has been updated."""

    schema_hash: str = Field(..., description="Schema hash after the update")

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.schema.update"

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.schema_branch.{self.branch}",
            "infrahub.branch.name": self.branch,
            "infrahub.branch.schema_hash": self.schema_hash,
        }

    def get_payload(self) -> dict[str, Any]:
        return {"branch": self.branch, "schema_hash": self.schema_hash}

    def get_messages(self) -> list[InfrahubMessage]:
        return [
            RefreshRegistryBranches(),
            # EventSchemaUpdate(
            #     branch=self.branch,
            #     meta=self.get_message_meta(),
            # )
        ]



from typing import Optional

from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase

from .interface import MutationNodeGetterInterface


class MutationNodeGetterByDefaultFilter(MutationNodeGetterInterface):
    def __init__(self, db: InfrahubDatabase, node_manager: NodeManager) -> None:
        self.db = db
        self.node_manager = node_manager

    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Optional[Node]:
        node = None
        default_filter_value = None
        if not node_schema.default_filter:
            return node
        this_datum = data

        for filter_key in node_schema.default_filter.split("__"):
            if filter_key not in this_datum:
                break
            this_datum = this_datum[filter_key]
        default_filter_value = this_datum

        if not default_filter_value:
            return node

        return await self.node_manager.get_one_by_default_filter(
            db=self.db,
            id=default_filter_value,
            kind=node_schema.kind,
            branch=branch,
        )



from infrahub_sdk.uuidt import UUIDT
from prefect import flow
from prefect.logging import get_run_logger

from infrahub.core.constants import InfrahubKind
from infrahub.core.timestamp import Timestamp
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices
from infrahub.workflows.utils import add_tags


@flow(name="repository-check", flow_run_name="Running repository checks for repository {message.repository}")
async def checks(message: messages.RequestRepositoryChecks, service: InfrahubServices) -> None:
    """Request to start validation checks on a specific repository."""
    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change])
    log = get_run_logger()

    events: list[InfrahubMessage] = []

    repository = await service.client.get(
        kind=InfrahubKind.GENERICREPOSITORY, id=message.repository, branch=message.source_branch
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=message.proposed_change)

    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []
    await proposed_change.validations.fetch()
    await repository.checks.fetch()

    validator_name = f"Repository Validator: {repository.name.value}"
    validator = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer

        if (
            existing_validator.typename == InfrahubKind.REPOSITORYVALIDATOR
            and existing_validator.repository.id == message.repository
        ):
            validator = existing_validator

    if validator:
        validator.conclusion.value = "unknown"
        validator.state.value = "queued"
        validator.started_at.value = ""
        validator.completed_at.value = ""
        await validator.save()
    else:
        validator = await service.client.create(
            kind=InfrahubKind.REPOSITORYVALIDATOR,
            data={
                "label": validator_name,
                "proposed_change": message.proposed_change,
                "repository": message.repository,
            },
        )
        await validator.save()

    check_execution_id = str(UUIDT())
    check_execution_ids.append(check_execution_id)
    log.info("Adding check for merge conflict")

    events.append(
        messages.CheckRepositoryMergeConflicts(
            validator_id=validator.id,
            validator_execution_id=validator_execution_id,
            check_execution_id=check_execution_id,
            proposed_change=message.proposed_change,
            repository_id=message.repository,
            repository_name=repository.name.value,
            source_branch=message.source_branch,
            target_branch=message.target_branch,
        )
    )

    checks_in_execution = ",".join(check_execution_ids)
    log.info(f"Checks in execution {checks_in_execution}")
    await service.cache.set(
        key=f"validator_execution_id:{validator_execution_id}:checks",
        value=checks_in_execution,
        expires=KVTTL.TWO_HOURS,
    )
    events.append(
        messages.FinalizeValidatorExecution(
            start_time=Timestamp().to_string(),
            validator_id=validator.id,
            validator_execution_id=validator_execution_id,
            validator_type=InfrahubKind.REPOSITORYVALIDATOR,
        )
    )

    for event in events:
        event.assign_meta(parent=message)
        await service.send(message=event)


@flow(
    name="repository-users-check",
    flow_run_name="Evaluating user-defined checks on repository {message.repository_name}",
)
async def user_checks(message: messages.RequestRepositoryUserChecks, service: InfrahubServices) -> None:
    """Request to start validation checks on a specific repository for User-defined checks."""

    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change])
    log = get_run_logger()

    events: list[InfrahubMessage] = []

    repository = await service.client.get(
        kind=InfrahubKind.GENERICREPOSITORY, id=message.repository_id, branch=message.source_branch, fragment=True
    )
    await repository.checks.fetch()

    for relationship in repository.checks.peers:
        log.info("Adding check for user defined check")
        check_definition = relationship.peer
        events.append(
            messages.CheckRepositoryCheckDefinition(
                check_definition_id=check_definition.id,
                repository_id=repository.id,
                repository_name=repository.name.value,
                commit=repository.commit.value,
                file_path=check_definition.file_path.value,
                class_name=check_definition.class_name.value,
                branch_name=message.source_branch,
                proposed_change=message.proposed_change,
                branch_diff=message.branch_diff,
            )
        )

    for event in events:
        event.assign_meta(parent=message)
        await service.send(message=event)



from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Any, Literal, Union

import ujson
import yaml
from pydantic import BaseModel, ConfigDict, Field

from .exceptions import DirectoryNotFoundError


class InfrahubTestExpectedResult(str, Enum):
    PASS = "PASS"
    FAIL = "FAIL"


class InfrahubTestResource(str, Enum):
    CHECK = "Check"
    JINJA2_TRANSFORM = "Jinja2Transform"
    PYTHON_TRANSFORM = "PythonTransform"
    GRAPHQL_QUERY = "GraphQLQuery"


class InfrahubBaseTest(BaseModel):
    """Basic Infrahub test model used as a common ground for all tests."""


class InfrahubInputOutputTest(InfrahubBaseTest):
    directory: Path | None = Field(
        None, description="Path to the directory where the input and output files are located"
    )
    input: Path = Field(
        Path("input.json"),
        description="Path to the file with the input data for the test, can be a relative path from the config file or from the directory.",
    )
    output: Path | None = Field(
        None,
        description="Path to the file with the expected output for the test, can be a relative path from the config file or from the directory.",
    )

    @staticmethod
    def parse_user_provided_data(path: Path | None) -> Any:
        """Read and parse user provided data depending on a file extension.

        This function handles JSON and YAML as they can be used to achieve the same goal. However some users may be more used to one format or
        another. If the file extension isn't known, assume the content is plain text.
        """
        if path is None:
            return None

        suffix = path.suffix.lower()[1:] if path.suffix else ""
        text = path.read_text()

        if suffix and suffix == "json":
            return ujson.loads(text)
        if suffix in ("yml", "yaml"):
            return yaml.safe_load(text)

        return text

    def update_paths(self, base_dir: Path) -> None:
        if self.directory and not self.directory.is_absolute() and not self.directory.is_dir():
            self.directory = Path(base_dir / self.directory)
            if not self.directory.is_dir():
                raise DirectoryNotFoundError(name=str(self.directory))
        else:
            self.directory = base_dir

        if not self.input or not self.input.is_file():
            search_input: Path | str = self.input or "input.*"
            results = list(self.directory.rglob(str(search_input)))

            if not results:
                raise FileNotFoundError(self.input)
            if len(results) != 1:
                raise FileNotFoundError(
                    f"Too many files are matching: {self.input}, please set the 'input' test key to the file to use."
                )
            self.input = results[0]

        if not self.output or not self.output.is_file():
            search_output: Path | str = self.output or "output.*"
            results = list(self.directory.rglob(str(search_output)))

            if results and len(results) != 1:
                raise FileNotFoundError(
                    f"Too many files are matching: {self.output}, please set the 'output' test key to the file to use."
                )
            if results:
                self.output = results[0]

    def get_input_data(self) -> Any:
        return self.parse_user_provided_data(self.input)

    def get_output_data(self) -> Any:
        return self.parse_user_provided_data(self.output)


class InfrahubIntegrationTest(InfrahubInputOutputTest):
    variables: Path | dict[str, Any] = Field(
        Path("variables.json"), description="Variables and corresponding values to pass to the GraphQL query"
    )

    def update_paths(self, base_dir: Path) -> None:
        super().update_paths(base_dir)

        if self.variables and not isinstance(self.variables, dict) and not self.variables.is_file():
            search_variables: Path | str = self.variables or "variables.*"
            results = list(self.directory.rglob(str(search_variables)))  # type: ignore[union-attr]

            if not results:
                raise FileNotFoundError(self.variables)
            if len(results) != 1:
                raise FileNotFoundError(
                    f"Too many files are matching: {self.variables}, please set the 'variables' test key to the file to use."
                )
            self.variables = results[0]

    def get_variables_data(self) -> dict[str, Any]:
        if isinstance(self.variables, dict):
            return self.variables
        return self.parse_user_provided_data(self.variables)


class InfrahubCheckSmokeTest(InfrahubBaseTest):
    kind: Literal["check-smoke"]


class InfrahubCheckUnitProcessTest(InfrahubInputOutputTest):
    kind: Literal["check-unit-process"]


class InfrahubCheckIntegrationTest(InfrahubIntegrationTest):
    kind: Literal["check-integration"]


class InfrahubGraphQLQuerySmokeTest(InfrahubBaseTest):
    kind: Literal["graphql-query-smoke"]
    path: Path = Field(..., description="Path to the file in which the GraphQL query is defined")


class InfrahubGraphQLQueryIntegrationTest(InfrahubIntegrationTest):
    kind: Literal["graphql-query-integration"]
    query: str = Field(..., description="Name of a pre-defined GraphQL query to execute")


class InfrahubJinja2TransformSmokeTest(InfrahubBaseTest):
    kind: Literal["jinja2-transform-smoke"]


class InfrahubJinja2TransformUnitRenderTest(InfrahubInputOutputTest):
    kind: Literal["jinja2-transform-unit-render"]


class InfrahubJinja2TransformIntegrationTest(InfrahubIntegrationTest):
    kind: Literal["jinja2-transform-integration"]


class InfrahubPythonTransformSmokeTest(InfrahubBaseTest):
    kind: Literal["python-transform-smoke"]


class InfrahubPythonTransformUnitProcessTest(InfrahubInputOutputTest):
    kind: Literal["python-transform-unit-process"]


class InfrahubPythonTransformIntegrationTest(InfrahubIntegrationTest):
    kind: Literal["python-transform-integration"]


class InfrahubTest(BaseModel):
    name: str = Field(..., description="Name of the test, must be unique")
    expect: InfrahubTestExpectedResult = Field(
        InfrahubTestExpectedResult.PASS,
        description="Expected outcome of the test, can be either PASS (default) or FAIL",
    )
    spec: Union[  # noqa: UP007
        InfrahubCheckSmokeTest,
        InfrahubCheckUnitProcessTest,
        InfrahubCheckIntegrationTest,
        InfrahubGraphQLQuerySmokeTest,
        InfrahubGraphQLQueryIntegrationTest,
        InfrahubJinja2TransformSmokeTest,
        InfrahubJinja2TransformUnitRenderTest,
        InfrahubJinja2TransformIntegrationTest,
        InfrahubPythonTransformSmokeTest,
        InfrahubPythonTransformUnitProcessTest,
        InfrahubPythonTransformIntegrationTest,
    ] = Field(..., discriminator="kind")


class InfrahubTestGroup(BaseModel):
    resource: InfrahubTestResource
    resource_name: str
    tests: list[InfrahubTest]


class InfrahubTestFileV1(BaseModel):
    model_config = ConfigDict(extra="forbid")
    version: str | None = "1.0"
    infrahub_tests: list[InfrahubTestGroup]



from typing import Union

from prefect import flow

from infrahub.core.constants import InfrahubKind, ValidatorConclusion
from infrahub.core.timestamp import Timestamp
from infrahub.git.repository import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.services import InfrahubServices
from infrahub.tasks.artifact import define_artifact
from infrahub.tasks.check import set_check_status
from infrahub.workflows.utils import add_tags

log = get_logger()


@flow(name="git-repository-check-artifact-create", flow_run_name="Check artifact creation")
async def create(message: messages.CheckArtifactCreate, service: InfrahubServices) -> None:
    await add_tags(branches=[message.branch_name], nodes=[message.target_id])
    validator = await service.client.get(
        kind=InfrahubKind.ARTIFACTVALIDATOR, id=message.validator_id, include=["checks"]
    )

    if InfrahubKind.READONLYREPOSITORY:
        repo = await InfrahubReadOnlyRepository.init(
            id=message.repository_id, name=message.repository_name, client=service.client
        )
    else:
        repo = await InfrahubRepository.init(
            id=message.repository_id, name=message.repository_name, client=service.client
        )

    artifact = await define_artifact(message=message, service=service)

    conclusion = ValidatorConclusion.SUCCESS.value
    severity = "info"
    artifact_result: dict[str, Union[str, bool, None]] = {
        "changed": None,
        "checksum": None,
        "artifact_id": None,
        "storage_id": None,
    }
    check_message = "Failed to render artifact"

    try:
        result = await repo.render_artifact(artifact=artifact, message=message)
        artifact_result["changed"] = result.changed
        artifact_result["checksum"] = result.checksum
        artifact_result["artifact_id"] = result.artifact_id
        artifact_result["storage_id"] = result.storage_id
        check_message = "Artifact rendered successfully"

    except Exception as exc:  # pylint: disable=broad-except
        conclusion = ValidatorConclusion.FAILURE.value
        artifact.status.value = "Error"
        severity = "critical"
        check_message += f": {str(exc)}"
        await artifact.save()

    check = None
    check_name = f"{message.artifact_name}: {message.target_name}"
    existing_check = await service.client.filters(
        kind=InfrahubKind.ARTIFACTCHECK, validator__ids=validator.id, name__value=check_name
    )
    if existing_check:
        check = existing_check[0]

    if check:
        check.created_at.value = Timestamp().to_string()
        check.conclusion.value = conclusion
        check.severity.value = severity
        check.changed.value = artifact_result["changed"]
        check.checksum.value = artifact_result["checksum"]
        check.artifact_id.value = artifact_result["artifact_id"]
        check.storage_id.value = artifact_result["storage_id"]
        await check.save()
    else:
        check = await service.client.create(
            kind=InfrahubKind.ARTIFACTCHECK,
            data={
                "name": check_name,
                "origin": message.repository_id,
                "kind": "ArtifactDefinition",
                "validator": message.validator_id,
                "created_at": Timestamp().to_string(),
                "message": check_message,
                "conclusion": conclusion,
                "severity": severity,
                "changed": artifact_result["changed"],
                "checksum": artifact_result["checksum"],
                "artifact_id": artifact_result["artifact_id"],
                "storage_id": artifact_result["storage_id"],
            },
        )
        await check.save()

    await set_check_status(message=message, conclusion=conclusion, service=service)



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventBranchMerge(InfrahubMessage):
    """Sent when a branch has been merged."""

    source_branch: str = Field(..., description="The source branch")
    target_branch: str = Field(..., description="The target branch")



from __future__ import annotations

from typing import Any, Sequence

from ..query import AttributeMigrationQuery
from ..query.attribute_add import AttributeAddQuery
from ..shared import AttributeSchemaMigration


class NodeAttributeAddMigrationQuery01(AttributeMigrationQuery, AttributeAddQuery):
    name = "migration_node_attribute_add_01"

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        super().__init__(
            migration=migration,
            node_kind=migration.new_schema.kind,
            attribute_name=migration.new_attribute_schema.name,
            attribute_kind=migration.new_attribute_schema.kind,
            branch_support=migration.new_attribute_schema.get_branch().value,
            default_value=migration.new_attribute_schema.default_value,
            **kwargs,
        )


class NodeAttributeAddMigration(AttributeSchemaMigration):
    name: str = "node.attribute.add"
    queries: Sequence[type[AttributeMigrationQuery]] = [NodeAttributeAddMigrationQuery01]  # type: ignore[assignment]



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import APIRouter, Body, Depends, Request, Response
from pydantic import BaseModel, Field

from infrahub.api.dependencies import BranchParams, get_branch_params, get_current_user, get_db, get_permission_manager
from infrahub.core import registry
from infrahub.core.account import ObjectPermission
from infrahub.core.constants import GLOBAL_BRANCH_NAME, InfrahubKind, PermissionAction
from infrahub.core.protocols import CoreArtifactDefinition
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.exceptions import NodeNotFoundError
from infrahub.git.models import RequestArtifactDefinitionGenerate
from infrahub.log import get_logger
from infrahub.permissions.constants import PermissionDecisionFlag
from infrahub.workflows.catalogue import REQUEST_ARTIFACT_DEFINITION_GENERATE

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.permissions import PermissionManager

log = get_logger()
router = APIRouter(prefix="/artifact")


class ArtifactGeneratePayload(BaseModel):
    nodes: list[str] = Field(default_factory=list)


class ArtifactGenerateResponse(BaseModel):
    nodes: list[str]


@router.get("/{artifact_id:str}")
async def get_artifact(
    artifact_id: str,
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    _: AccountSession = Depends(get_current_user),
) -> Response:
    artifact = await registry.manager.get_one(db=db, id=artifact_id, branch=branch_params.branch, at=branch_params.at)
    if not artifact:
        raise NodeNotFoundError(
            branch_name=branch_params.branch.name, node_type=InfrahubKind.ARTIFACT, identifier=artifact_id
        )

    return Response(
        content=registry.storage.retrieve(identifier=artifact.storage_id.value),
        headers={"Content-Type": artifact.content_type.value.value},
    )


@router.post("/generate/{artifact_definition_id:str}")
async def generate_artifact(
    request: Request,
    artifact_definition_id: str,
    payload: ArtifactGeneratePayload = Body(
        ArtifactGeneratePayload(),
        description="Payload of the request, can be used to limit the scope of the query to a specific list of hosts",
    ),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    permission_manager: PermissionManager = Depends(get_permission_manager),
) -> None:
    permission_decision = (
        PermissionDecisionFlag.ALLOW_DEFAULT
        if branch_params.branch.name in (GLOBAL_BRANCH_NAME, registry.default_branch)
        else PermissionDecisionFlag.ALLOW_OTHER
    )
    permissions = [
        ObjectPermission(namespace="Core", name="Artifact", action=action.value, decision=permission_decision)
        for action in (PermissionAction.CREATE, PermissionAction.UPDATE)
    ]
    permission_manager.raise_for_permissions(permissions=permissions)

    # Verify that the artifact definition exists for the requested branch
    artifact_definition = await registry.manager.get_one_by_id_or_default_filter(
        db=db,
        id=artifact_definition_id,
        kind=CoreArtifactDefinition,
        branch=branch_params.branch,
    )

    service = request.app.state.service
    model = RequestArtifactDefinitionGenerate(
        artifact_definition_id=artifact_definition.id,
        artifact_definition_name=artifact_definition.name.value,
        branch=branch_params.branch.name,
        limit=payload.nodes,
    )

    await service.workflow.submit_workflow(workflow=REQUEST_ARTIFACT_DEFINITION_GENERATE, parameters={"model": model})



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

import typer
from rich.console import Console

from ..ctl import config
from ..ctl.client import initialize_client
from ..ctl.repository import get_repository_config
from ..ctl.utils import execute_graphql_query, init_logging, parse_cli_vars
from ..exceptions import ModuleImportError
from ..node import InfrahubNode

if TYPE_CHECKING:
    from ..schema.repository import InfrahubRepositoryConfig


async def run(
    generator_name: str,
    path: str,  # noqa: ARG001
    debug: bool,
    list_available: bool,
    branch: str | None = None,
    variables: Optional[list[str]] = None,
) -> None:
    init_logging(debug=debug)
    repository_config = get_repository_config(Path(config.INFRAHUB_REPO_CONFIG_FILE))

    if list_available or not generator_name:
        list_generators(repository_config=repository_config)
        return

    generator_config = repository_config.get_generator_definition(name=generator_name)

    console = Console()
    relative_path = str(generator_config.file_path.parent) if generator_config.file_path.parent != Path() else None

    try:
        generator_class = generator_config.load_class(import_root=str(Path.cwd()), relative_path=relative_path)
    except ModuleImportError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    variables_dict = parse_cli_vars(variables)

    param_key = list(generator_config.parameters.keys())
    identifier = None
    if param_key:
        identifier = param_key[0]

    client = initialize_client(branch=branch)
    if variables_dict:
        data = execute_graphql_query(
            query=generator_config.query,
            variables_dict=variables_dict,
            branch=branch,
            debug=False,
            repository_config=repository_config,
        )
        generator = generator_class(
            query=generator_config.query,
            client=client,
            branch=branch,
            params=variables_dict,
            convert_query_response=generator_config.convert_query_response,
            infrahub_node=InfrahubNode,
        )
        await generator._init_client.schema.all(branch=generator.branch_name)
        await generator.run(identifier=generator_config.name, data=data)

    else:
        targets = await client.get(
            kind="CoreGroup", branch=branch, include=["members"], name__value=generator_config.targets
        )
        await targets.members.fetch()

        if not targets.members.peers:
            console.print(
                f"[red]No members found within '{generator_config.targets}', not running generator '{generator_name}'"
            )
            return

        for member in targets.members.peers:
            check_parameter = {}
            if identifier:
                attribute = getattr(member.peer, identifier)
                check_parameter = {identifier: attribute.value}
            params = {"name": member.peer.name.value}
            generator = generator_class(
                query=generator_config.query,
                client=client,
                branch=branch,
                params=params,
                convert_query_response=generator_config.convert_query_response,
                infrahub_node=InfrahubNode,
            )
            data = execute_graphql_query(
                query=generator_config.query,
                variables_dict=check_parameter,
                branch=branch,
                debug=False,
                repository_config=repository_config,
            )
            await generator._init_client.schema.all(branch=generator.branch_name)
            await generator.run(identifier=generator_config.name, data=data)


def list_generators(repository_config: InfrahubRepositoryConfig) -> None:
    console = Console()
    console.print(f"Generators defined in repository: {len(repository_config.generator_definitions)}")

    for generator in repository_config.generator_definitions:
        console.print(f"{generator.name} ({generator.file_path}::{generator.class_name}) Target: {generator.targets}")
